# -*- coding: utf-8 -*-
from marshmallow import Schema, fields

__author__ = "aganezov"
__email__ = "aganezov(at)gwu.edu"
__status__ = "production"

# module wide constant, that is utilized for json dict key creation
BGGenome_JSON_SCHEMA_JSON_KEY = "_py__bg_genome_json_schema"


class BGGenome(object):
    """ A class that represent a genome object for the breakpoint graph

     For purposes of breakpoint graph no additional information about genome is needed, except its name, that is used in various
     algorithmic tasks (multicolor splitting, tree traversing, etc)
    """
    class BGGenomeJSONSchema(Schema):
        """ a JSON schema powered by marshmallow library to serialize/deserialize genome object into/from JSON format
        """
        _py__bg_genome_json_schema = fields.String(attribute="json_schema_name")
        g_id = fields.Integer(attribute="json_id")
        name = fields.String()

        def make_object(self, data):
            if "name" not in data:
                raise ValueError("Error during genome serialization. \"name\" key is not present in json object")
            name = data["name"]
            return BGGenome(name=name)

    # class wide variable for json serialization/deserialization. Created once for a whole class, as thousands of objects
    # undergo serialization / deserialization, and schema instantiation in each case would require additional resources
    json_schema = BGGenomeJSONSchema()

    def __init__(self, name):
        self.name = name

    def __eq__(self, other):
        """ Two genomes a called equal if they are of same class and their hash values are equal to each other """
        if not isinstance(other, BGGenome):
            return False
        return hash(self) == hash(other)

    def __hash__(self):
        """ Since for breakpoint graph purposes distinction between genomes is made purely by their name, hash value of genome
         is proxied to hash value of genomes name
         """
        return hash(self.name)

    @property
    def json_id(self):
        """ A genome is references multiple times, as for example in multicolor object, and such reference is done by genome
        unique json id.
        """
        return hash(self)

    @property
    def json_schema_name(self):
        """ When genome is serialized information about JSON schema of such serialization can be recorded,
        and this property provides access to it
        """
        return self.json_schema.__class__.__name__

    def to_json(self, schema_info=True):
        """ JSON serialization method that accounts for a possibility of field filtration and schema specification """
        old_exclude_fields = self.json_schema.exclude
        new_exclude_fields = list(old_exclude_fields)
        if not schema_info:
            new_exclude_fields.append(BGGenome_JSON_SCHEMA_JSON_KEY)
        # monkey patch schema `exclude` attribute to ignore some fields in result json object
        self.json_schema.exclude = new_exclude_fields
        result = self.json_schema.dump(self).data
        # reverse the result of monkey patching
        self.json_schema.exclude = old_exclude_fields
        return result

    @classmethod
    def from_json(cls, data, json_schema_class=None):
        """ JSON deserialization method that retrieves a genome instance from its json representation

        If specific json schema is provided, it is utilized, and if not, a class specific is used
        """
        schema = cls.json_schema if json_schema_class is None else json_schema_class()
        return schema.load(data).data

    def __lt__(self, other):
        """ Genomes are ordered according to lexicographical ordering of their names """
        if not isinstance(other, BGGenome):
            return True
        return self.name < other.name

    def __le__(self, other):
        """ Genomes are ordered according to lexicographical ordering of their names """
        return self < other or self == other


# -*- coding: utf-8 -*-
__author__ = "Sergey Aganezov"
__email__ = "aganezov(at)gwu.edu"
__status__ = "production"

import unittest
from bg.vertices import BGVertex_JSON_SCHEMA_JSON_KEY, BGVertex, BlockVertex, InfinityVertex, TaggedVertex, TaggedBlockVertex, \
    TaggedInfinityVertex


class BGVertexTestCase(unittest.TestCase):
    def setUp(self):
        """ Heavily utilized values in multiple test """
        self.str_name1 = "name1"
        self.str_name2 = "name2"
        self.str_name3 = "name3"
        self.str_name4 = "name4"
        self.str_name5 = "name5"
        self.int_name1 = 1
        self.int_name2 = 2
        self.int_name3 = 3
        self.vertex_class = BGVertex

    def test_empty_initialization(self):
        with self.assertRaises(TypeError):
            self.vertex_class()

    def test_initialization(self):
        v = self.vertex_class(self.str_name1)
        self.assertEqual(v.name, self.str_name1)

    def test__hash__(self):
        """ hash value from BGVertex is defines as hash value from the name of respective vertex """
        self.assertEqual(hash(self.vertex_class(self.str_name1)), hash(self.vertex_class(self.str_name1).name))

    def test__eq__(self):
        """ BGVertices are equal is they are of the same class and have equal hash values """
        v1 = self.vertex_class(self.str_name1)
        v2 = self.vertex_class(self.str_name1)
        v3 = self.vertex_class(self.str_name3)
        self.assertEqual(v1, v2)
        self.assertNotEqual(v1, v3)
        self.assertNotEqual(v2, v3)
        for non_bg_vertex_value in [1, "1", (1,), [1]]:
            self.assertNotEqual(v1, non_bg_vertex_value)

    def test_is_something_vertex_request_stopper(self):
        """ BGVertex serves as a "stopper" class for all lookups like is_.+_vertex and returns False to all of them
            Examples:
                vertex.is_regular_vertex
                vertex.is_infinity_vertex
                vertex.is_block_vertex
        """
        for vertex_type_string in ["lala1", "lala2", "lala2", "lala4", "etc"]:
            # dummy values have no meaning, so inherited test cases can recall this test method
            vertex_request_type = "is_" + vertex_type_string + "_vertex"
            self.assertFalse(getattr(self.vertex_class(self.str_name1), vertex_request_type))

    def test_json_identifier(self):
        # json identifier shall be unique per BGVertex and be of `int` type
        # json identifier is implemented as a property on BGVertex object
        # __hash__ is used
        v = self.vertex_class(self.str_name1)
        json_id = v.json_id
        self.assertTrue(isinstance(json_id, int))
        self.assertEqual(json_id, hash(v))
        v.name = self.int_name2
        new_json_id = v.json_id
        self.assertTrue(isinstance(new_json_id, int))
        self.assertEqual(new_json_id, hash(v))
        self.assertNotEqual(json_id, new_json_id)

    def test_json_serialization(self):
        # a BGVertex class instance shall be serializable into the JSON based object
        # such serialization shall contain all breakpoint graph relative information about the vertex
        # as well as a special json_id field that unique identifies this instance of BGVertex
        v = self.vertex_class(self.str_name1)
        ref_result = {
            "name": self.str_name1,
            'v_id': v.json_id
        }
        result = v.to_json(schema_info=False)
        self.assertDictEqual(ref_result, result)
        # with json_schema_key
        ref_result[BGVertex_JSON_SCHEMA_JSON_KEY] = v.json_schema_name
        result = v.to_json(schema_info=True)
        self.assertDictEqual(ref_result, result)
        # a non-string name attribute value shall be translated into a string object
        v = self.vertex_class(self.int_name1)
        ref_result = {
            "name": "1",
            'v_id': v.json_id
        }
        result = v.to_json(schema_info=False)
        self.assertDictEqual(ref_result, result)

    def test_json_deserialization_correct_default_schema(self):
        # a BGVertex class instance is to be read from the JSON based object data
        # not all JSON object attributes are to be stored in BGVertex object
        json_object = {
            "name": self.str_name1,
            "_py__bg_vertex_json_schema": self.vertex_class.json_schema_name,
            "json_id": 1
        }
        vertex = self.vertex_class.from_json(json_object)
        self.assertIsInstance(vertex, self.vertex_class)
        # explicitly specified json schema to work with
        # but since nothing is passed into in "schema" argument to the from_json function
        # default to the BGVertex.json_schema is performed
        json_schema_name = self.vertex_class(self.str_name1).json_schema_name
        json_object = {
            BGVertex_JSON_SCHEMA_JSON_KEY: json_schema_name,
            "name": self.str_name1,
            "json_id": 1
        }
        vertex = self.vertex_class.from_json(json_object)
        self.assertIsInstance(vertex, self.vertex_class)
        self.assertEqual(vertex.name, self.str_name1)
        # there must be a "name" field in json object to be serialized
        with self.assertRaises(ValueError):
            json_object = {"json_id": 1}
            self.vertex_class.from_json(json_object)

    def test_get_vertex_class_from_vertex_name(self):
        # correct versions
        # as different types of vertices have different repr names,
        # they shall be ambiguously identified by their string name
        block_vertex_class = TaggedBlockVertex
        infinity_vertex_class = TaggedInfinityVertex
        bv = TaggedBlockVertex(self.str_name1)
        iv = TaggedInfinityVertex(self.str_name1)
        self.assertEqual(BGVertex.get_vertex_class_from_vertex_name(bv.name), block_vertex_class)
        self.assertEqual(BGVertex.get_vertex_class_from_vertex_name(iv.name), infinity_vertex_class)

    def test_get_vertex_name_root(self):
        # strips all suffixes from the string name and returns the root of it
        # most vertices, except for a BlockVertex will have single of multiple suffixes to it
        #   Example:
        #       InfinityVertex("lala").name == lala__infinity
        # "__" serves as name_separator, and is inherited from BGVertex. Everything before the first __ is considered
        # vertex name root
        string = "aaa"
        self.assertEqual(BGVertex.get_vertex_name_root(string), "aaa")
        string = ""
        self.assertEqual(BGVertex.get_vertex_name_root(string), "")
        string = "aaa" + BGVertex.NAME_SEPARATOR + ""
        self.assertEqual(BGVertex.get_vertex_name_root(string), "aaa")
        string = "aaa" + BGVertex.NAME_SEPARATOR + "aaa" + BGVertex.NAME_SEPARATOR + "aaa"
        self.assertEqual(BGVertex.get_vertex_name_root(string), "aaa")


class BlockVertexTestCase(BGVertexTestCase):
    """ Update vertex_class to call BlockVertex rather than BGVertex, and update overwritten portions of the class """

    def setUp(self):
        super().setUp()
        self.vertex_class = BlockVertex

    def test_inheritance(self):
        self.assertIsInstance(self.vertex_class(self.str_name1), BGVertex)

    def test_is_block_vertex(self):
        # BlockVertex defines two properties
        #   * is_regular_vertex, which is a broad class of vertices, that correspond to ends of fully present genomic blocks
        #   * is_block_vertex, which is a narrower class of vertices, that correspond to ends of gene/synteny blocks
        v = self.vertex_class(self.str_name1)
        self.assertTrue(v.is_regular_vertex)
        self.assertTrue(v.is_block_vertex)

    def test_is_head_vertex(self):
        self.assertTrue(self.vertex_class("vertexh").is_head_vertex)
        self.assertFalse(self.vertex_class("vertexl").is_head_vertex)

    def test_is_tail_vertex(self):
        self.assertFalse(self.vertex_class("vertexl").is_tail_vertex)
        self.assertTrue(self.vertex_class("vertext").is_tail_vertex)

    def test_get_vertex_mate_if_no_specified(self):
        v = self.vertex_class(self.str_name1)
        self.assertIsNone(v.mate_vertex)

    def test_get_vertex_mate(self):
        v = self.vertex_class(self.str_name1)
        v1 = self.vertex_class(self.str_name2)
        v.mate_vertex = v1
        v1.mate_vertex = v
        self.assertEqual(v.mate_vertex, v1)
        self.assertEqual(v1.mate_vertex, v)

    def test_block_name(self):
        self.assertEqual(self.vertex_class("vertexh").block_name, "vertex")
        self.assertEqual(self.vertex_class("vertext").block_name, "vertex")
        self.assertEqual(self.vertex_class("vertex").block_name, "vertex")


class InfinityVertexTestCase(BGVertexTestCase):
    """ Update vertex_class to call InfinityVertex rather than BGVertex, and update overwritten portions of the class """

    def setUp(self):
        super().setUp()
        self.vertex_class = InfinityVertex
        self.block_vertex = BlockVertex(self.str_name1)

    def test_initialization(self):
        #
        # InfinityVertex overwrites the access to the "name" attribute, thus keeping internal representation as is,
        # while returning on call a calculated on the fly value, that has InfinityVertex name suffix in it
        i_v = InfinityVertex(self.block_vertex.name)
        ref_name = InfinityVertex.NAME_SEPARATOR.join([self.block_vertex.name, InfinityVertex.NAME_SUFFIX])
        self.assertEqual(i_v.name, ref_name)

    def test_is_irregular_vertex(self):
        # InfinityVertex defines two properties
        #   * is_irregular_vertex, which is a broad class of vertices, that correspond to ends of genomic fragments
        #   * is_infinity__vertex, which is a narrower class of vertices, that correspond to traditional extrimities of genomic fragments
        i_v = InfinityVertex(self.block_vertex)
        self.assertTrue(i_v.is_infinity_vertex)
        self.assertTrue(i_v.is_irregular_vertex)

    def test_is_head_vertex(self):
        self.assertFalse(self.vertex_class("vertexh").is_tail_vertex)

    def test_is_tail_vertex(self):
        self.assertFalse(self.vertex_class("vertext").is_tail_vertex)

    def test_inheritance(self):
        self.assertIsInstance(InfinityVertex(self.block_vertex), BGVertex)

    def test_json_identifier(self):
        # json identifier shall be unique per BGVertex and be of `int` type
        # json identifier is implemented as a property on BGVertex object
        # __hash__ is used
        v = self.vertex_class(self.block_vertex.name)
        json_id = v.json_id
        self.assertTrue(isinstance(json_id, int))
        self.assertEqual(json_id, hash(v))
        v.name = self.int_name2
        new_json_id = v.json_id
        self.assertTrue(isinstance(new_json_id, int))
        self.assertEqual(new_json_id, hash(v))
        self.assertNotEqual(json_id, new_json_id)

    def test_json_serialization(self):
        # a InfinityVertex class instance shall be serializable into the JSON based object
        # such serialization shall contain all breakpoint graph relative information about the vertex
        # as well as a special json_id field that unique identifies this instance of BGVertex
        v = self.vertex_class(self.block_vertex.name)
        ref_result = {
            "name": InfinityVertex.NAME_SEPARATOR.join([self.str_name1, InfinityVertex.NAME_SUFFIX]),
            'v_id': v.json_id
        }
        result = v.to_json(schema_info=False)
        self.assertDictEqual(ref_result, result)
        # with json_schema_key
        ref_result[BGVertex_JSON_SCHEMA_JSON_KEY] = v.json_schema_name
        result = v.to_json(schema_info=True)
        self.assertDictEqual(ref_result, result)
        # a non-string name attribute value shall be translated into a string object
        v = self.vertex_class(BlockVertex(1).name)
        ref_result = {
            "name": InfinityVertex.NAME_SEPARATOR.join(["1", InfinityVertex.NAME_SUFFIX]),
            'v_id': v.json_id
        }
        result = v.to_json(schema_info=False)
        self.assertDictEqual(ref_result, result)

    def test_json_deserialization_correct_default_schema(self):
        # a BGVertex class instance is to be read from the JSON based object data
        # not all JSON object attributes are to be stored in BGVertex object
        json_object = {
            "name": self.str_name1,
            "json_id": 1
        }
        vertex = self.vertex_class.from_json(json_object)
        self.assertTrue(isinstance(vertex, self.vertex_class))
        # explicitly specified json schema to work with
        # but since nothing is passed into in "schema" argument to the from_json function
        # default to the BGVertex.json_schema is performed
        json_schema_name = self.vertex_class(self.block_vertex).json_schema_name
        json_object = {
            BGVertex_JSON_SCHEMA_JSON_KEY: json_schema_name,
            "name": self.str_name1,
            "json_id": 1
        }
        vertex = self.vertex_class.from_json(json_object)
        self.assertTrue(isinstance(vertex, self.vertex_class))
        self.assertEqual(vertex.name, InfinityVertex.NAME_SEPARATOR.join([self.str_name1, InfinityVertex.NAME_SUFFIX]))
        # there must be a "name" field in json object to be serialized
        with self.assertRaises(ValueError):
            json_object = {"json_id": 1}
            self.vertex_class.from_json(json_object)


class TaggedVertexTestCase(BGVertexTestCase):
    """ Update vertex_class to call InfinityVertex rather than BGVertex, and update overwritten portions of the class """

    def setUp(self):
        super().setUp()
        self.vertex_class = TaggedVertex
        self.tagged_vertex = TaggedVertex(self.str_name1)

    def test_initialization(self):
        #
        # Tagged vertex overwrites access to the "name" attribute, thus keeping hte internal representation as is,
        # while returning on call a calculated on the fly value that has pairs of "tag" -- "value" in it
        t_v = self.vertex_class(self.str_name1)
        self.assertTrue(t_v.is_tagged_vertex)
        self.assertListEqual(t_v.tags, [])
        # when the tags list is empty the "name", when accessed, will be equal to the root portion of itself
        self.assertIn(self.str_name1, t_v.name)

    def test_is_tagged_vertex(self):
        self.assertTrue(self.tagged_vertex.is_tagged_vertex)

    def test_inheritance(self):
        self.assertIsInstance(self.vertex_class(self.str_name1), BGVertex)

    def test_add_tag(self):
        # tags in vertex are represented in pairs "tag" -- "value"
        # and suppose to be added in a respective way
        # order of tags is preserved throughout the adding process
        t_v = self.vertex_class(self.str_name1)
        t_v.add_tag("repeat", 1)
        self.assertIn(("repeat", 1), t_v.tags)

        # unique "tag" -- "value" pairs are to be present exactly once
        t_v.add_tag("repeat", 1)
        self.assertEqual(t_v.tags.count(("repeat", 1)), 1)

        # addition of a different tag with same name, but different value shall work just fine and as expected
        t_v.add_tag("repeat", 2)
        self.assertIn(("repeat", 2), t_v.tags)

        # added tags shall preserve the sorted order of the tags list
        t_v.add_tag("repeat", 0)
        self.assertLess(t_v.tags.index(("repeat", 0)), t_v.tags.index(("repeat", 1)))
        self.assertLess(t_v.tags.index(("repeat", 0)), t_v.tags.index(("repeat", 2)))

    def test_remove_tag(self):
        #
        # order of tags shall be preserved throughout deletion
        t_v = self.vertex_class(self.str_name1)
        t_v.add_tag("repeat", 2)
        t_v.add_tag("repeat", 3)
        t_v.add_tag("repeat", 0)
        t_v.add_tag("repeat", 1)

        length_before = len(t_v.tags)

        t_v.remove_tag("repeat", 2)

        length_after = len(t_v.tags)

        self.assertEqual(length_before - 1, length_after)

        self.assertLess(t_v.tags.index(("repeat", 0)), t_v.tags.index(("repeat", 1)))
        self.assertLess(t_v.tags.index(("repeat", 0)), t_v.tags.index(("repeat", 3)))
        self.assertLess(t_v.tags.index(("repeat", 1)), t_v.tags.index(("repeat", 3)))

        # silent fail options can be specified to remove the non present tag without raising the exception
        t_v.remove_tag("repeat", 2, silent_fail=True)

    def test_remove_tag_incorrect(self):
        # when attempting to remove a non existing "tag" -- "value" pair and no silent_fail option is present,
        #   a value error is raised
        with self.assertRaises(ValueError):
            self.tagged_vertex.remove_tag("lalala", "lalala")

    def test_tag_presence_in_name(self):
        #
        # for each "tag" -- "value" their str representation is used and is separated by ":" sign
        # example:
        #   tag   = "repeat"
        #   value = 1
        #   representation = "repeat:1"
        # multiple tags a separated between each other by the NAME_SEPARATOR
        t_v = self.vertex_class(self.str_name1)
        t_v.add_tag("repeat", 2)
        t_v.add_tag("repeat", 3)
        t_v.add_tag("repeat", 0)
        t_v.add_tag("repeat", 1)
        tags_as_strings = [self.tagged_vertex.TAG_SEPARATOR.join([str(tag), str(value)]) for tag, value in t_v.tags]
        ref_name = self.tagged_vertex.NAME_SEPARATOR.join([self.str_name1] + tags_as_strings)
        self.assertIn(ref_name,t_v.name)

    def test_is_something_vertex_based_on_tags_name(self):
        # all calls "is_something_vertex" shall be aware of the tags, that are stored in the vertex.tags container
        # if there is a tag -- key pair in the container, such, that tag equals to "something", the vertex shall be
        # be considered as "is_something_vertex"
        t_v = self.vertex_class(self.str_name1)
        tag_name = "tag_name_1"
        tag_value = 1
        t_v.add_tag(tag_name, tag_value)
        self.assertTrue(getattr(t_v, "is_" + tag_name + "_vertex"))
        # if the "something" is not found in tags, than the call shall be proxied forward
        self.assertTrue(t_v.is_tagged_vertex)
        # if the tag is not present, than the result of proxied call shall be returned
        t_v.remove_tag(tag_name, tag_value)
        self.assertFalse(getattr(t_v, "is_" + tag_name + "_vertex"))
        t_v.add_tag("tag", 1)
        t_v.add_tag("repeat", 1)
        self.assertTrue(t_v.is_repeat_vertex)
        t_v.remove_tag("repeat", 1)
        self.assertFalse(t_v.is_repeat_vertex)


class TaggedBlockVertexTestCase(TaggedVertexTestCase, BlockVertexTestCase):
    def setUp(self):
        super().setUp()
        self.vertex_class = TaggedBlockVertex
        self.tagged_block_vertex = TaggedBlockVertex(self.str_name1)

    def test_is_regular_and_is_tagged_vertex(self):
        self.assertTrue(self.tagged_block_vertex.is_regular_vertex)
        self.assertTrue(self.tagged_block_vertex.is_tagged_vertex)

    def test_name_creation_order(self):
        # name has to go as root + tags, just as in the tagged vertex
        tbv = self.vertex_class(self.str_name1)
        tbv.add_tag("tag1", 1)
        ref_name = BGVertex.NAME_SEPARATOR.join([self.str_name1] + [TaggedVertex.TAG_SEPARATOR.join([str(tag), str(value)])
                                                                    for tag, value in tbv.tags])
        self.assertIn(ref_name, tbv.name)

    def test_is_head_vertex(self):
        tbv = self.vertex_class("vertexh")
        tbv.add_tag("tag1", 1)
        self.assertTrue(tbv.is_head_vertex)

    def test_is_tail_vertex(self):
        tbv = self.vertex_class("vertext")
        tbv.add_tag("tag1", 1)
        self.assertTrue(tbv.is_tail_vertex)


class TaggedInfinityVertexTestCase(TaggedVertexTestCase, InfinityVertexTestCase):
    def setUp(self):
        super().setUp()
        self.vertex_class = TaggedInfinityVertex
        self.tagged_infinity_vertex = TaggedInfinityVertex(self.str_name1)

    def test_is_irregular_and_is_tagged_vertex(self):
        self.assertTrue(self.tagged_infinity_vertex.is_irregular_vertex)
        self.assertTrue(self.tagged_infinity_vertex.is_infinity_vertex)
        self.assertTrue(self.tagged_infinity_vertex.is_tagged_vertex)

    def test_name_creation_order(self):
        # name has to go as root + tags + infinity suffix
        tbv = self.vertex_class(self.str_name1)
        tbv.add_tag("tag1", 1)
        ref_name = BGVertex.NAME_SEPARATOR.join([self.str_name1] +
                                                [TaggedVertex.TAG_SEPARATOR.join([str(tag), str(value)]) for tag, value in tbv.tags] +
                                                [InfinityVertex.NAME_SUFFIX])
        self.assertIn(ref_name, tbv.name)

        # with no tag, just a single name separator is to be inserted between root, and infinity suffix
        tbv = self.vertex_class(self.str_name1)
        ref_name = BGVertex.NAME_SEPARATOR.join([self.str_name1] + [InfinityVertex.NAME_SUFFIX])
        self.assertIn(ref_name, tbv.name)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()  # pragma: no cover



# -*- coding: utf-8 -*-
from collections import Counter
from unittest.mock import Mock
from bg.genome import BGGenome
from bg.multicolor import Multicolor
from bg.vertices import BlockVertex, InfinityVertex

__author__ = "Sergey Aganezov"
__email__ = "aganezov(at)gwu.edu"
__status__ = "production"

import unittest
from bg.edge import BGEdge, BGEdge_JSON_SCHEMA_JSON_KEY


class BGEdgeTestCase(unittest.TestCase):

    def setUp(self):
        # heavily utilized variables
        self.genome1 = BGGenome("red")
        self.genome2 = BGGenome("green")
        self.genome3 = BGGenome("blue")
        self.genome4 = BGGenome("black")
        self.genome4 = BGGenome("yellow")

    def test_empty_initialization_incorrect(self):
        with self.assertRaises(TypeError):
            # a BGEdge wrapper is meant to wrap something, but not nothing
            BGEdge()

    def test_initialization(self):
        # simple correct initialization of BGEdge instance
        v1 = BlockVertex("v1")
        v2 = BlockVertex("v2")
        multicolor = Multicolor(self.genome3)
        edge = BGEdge(vertex1=v1,
                      vertex2=v2,
                      multicolor=multicolor)
        self.assertEqual(edge.vertex1, v1)
        self.assertEqual(edge.vertex2, v2)
        self.assertEqual(edge.multicolor, multicolor)

    def test_merging_correct(self):
        # two BGEdges can be merged together into a third, separate BGEdge
        # that would contain information from both supplied BGEdges in terms of colors and multiplicities
        # such merge is allowed only if a pair of vertices in both BGEdges is the same
        # ordering of vertices if not a concern, since edges in BreakpointGraph are not directed
        v1 = BlockVertex("v1")
        v2 = BlockVertex("v2")
        multicolor = Multicolor(self.genome3)
        multicolor1 = Multicolor(self.genome2)
        edge1 = BGEdge(vertex1=v1, vertex2=v2, multicolor=multicolor)
        edge2 = BGEdge(vertex1=v1, vertex2=v2, multicolor=multicolor1)
        merged_edge = BGEdge.merge(edge1, edge2)
        self.assertEqual(merged_edge.vertex1, v1)
        self.assertEqual(merged_edge.vertex2, v2)
        self.assertEqual(merged_edge.multicolor, multicolor + multicolor1)

    def test_merging_incorrect(self):
        # cases when vertices in two supplied for the merging edges are not consistent
        v1 = BlockVertex("v1")
        v2 = BlockVertex("v2")
        v3 = BlockVertex("v3")
        v4 = BlockVertex("v4")
        multicolor = Multicolor(self.genome3)
        multicolor1 = Multicolor(self.genome2)
        edge1 = BGEdge(vertex1=v1, vertex2=v2, multicolor=multicolor)
        edge2 = BGEdge(vertex1=v1, vertex2=v3, multicolor=multicolor1)
        with self.assertRaises(ValueError):
            BGEdge.merge(edge1, edge2)
        edge2 = BGEdge(vertex1=v3, vertex2=v2, multicolor=multicolor1)
        with self.assertRaises(ValueError):
            BGEdge.merge(edge1, edge2)
        edge2 = BGEdge(vertex1=v3, vertex2=v4, multicolor=multicolor1)
        with self.assertRaises(ValueError):
            BGEdge.merge(edge1, edge2)
        edge2 = BGEdge(vertex1=v1, vertex2=v1, multicolor=multicolor1)
        with self.assertRaises(ValueError):
            BGEdge.merge(edge1, edge2)
        edge2 = BGEdge(vertex1=v2, vertex2=v2, multicolor=multicolor1)
        with self.assertRaises(ValueError):
            BGEdge.merge(edge1, edge2)
        edge2 = BGEdge(vertex1=v3, vertex2=v1, multicolor=multicolor1)
        with self.assertRaises(ValueError):
            BGEdge.merge(edge1, edge2)

    def test_equality(self):
        # edges are called equal if they connect same pairs of vertices and have same multicolor assigned to them
        v1 = BlockVertex("v1")
        v2 = BlockVertex("v2")
        v3 = BlockVertex("v3")
        v4 = BlockVertex("v4")
        multicolor = Multicolor(self.genome3)
        multicolor1 = Multicolor(self.genome2)
        edge1 = BGEdge(vertex1=v1, vertex2=v2, multicolor=multicolor)
        edge2 = BGEdge(vertex1=v1, vertex2=v2, multicolor=multicolor1)
        edge3 = BGEdge(vertex1=v1, vertex2=v3, multicolor=multicolor1)
        edge4 = BGEdge(vertex1=v3, vertex2=v4, multicolor=multicolor)
        self.assertNotEqual(edge1, edge2)
        self.assertNotEqual(edge1, edge3)
        self.assertNotEqual(edge2, edge3)
        self.assertNotEqual(edge1, edge4)
        edge4 = BGEdge(vertex1=v2, vertex2=v1, multicolor=multicolor)
        edge5 = BGEdge(vertex1=v1, vertex2=v2, multicolor=multicolor)
        self.assertEqual(edge1, edge4)
        self.assertEqual(edge1, edge5)
        self.assertEqual(edge4, edge5)
        self.assertNotEqual(edge1, 5)
        edge6 = BGEdge(vertex1=v3, vertex2=v1, multicolor=multicolor)
        self.assertNotEqual(edge1, edge6)

    def test_is_infinity_edge(self):
        # and edge is called an infinity edge if at least one of its vertices is an infinity vertex
        v1 = BlockVertex("v1")
        v2 = BlockVertex("v2")
        v3 = InfinityVertex("v3")
        multicolor = Multicolor(self.genome3)
        edge1 = BGEdge(vertex1=v1, vertex2=v2, multicolor=multicolor)
        edge2 = BGEdge(vertex1=v1, vertex2=v3, multicolor=multicolor)
        edge3 = BGEdge(vertex1=v3, vertex2=v1, multicolor=multicolor)
        edge4 = BGEdge(vertex1=v3, vertex2=v3, multicolor=multicolor)
        self.assertFalse(edge1.is_infinity_edge)
        self.assertTrue(edge2.is_infinity_edge)
        self.assertTrue(edge3.is_infinity_edge)
        self.assertTrue(edge4.is_infinity_edge)

    def test_is_irregular_edge(self):
        # and edge is called an irregular edge if at least one of its vertices is an irregular vertex
        v1 = BlockVertex("v1")
        v2 = BlockVertex("v2")
        v3 = InfinityVertex("v3")
        multicolor = Multicolor(self.genome3)
        edge1 = BGEdge(vertex1=v1, vertex2=v2, multicolor=multicolor)
        edge2 = BGEdge(vertex1=v1, vertex2=v3, multicolor=multicolor)
        edge3 = BGEdge(vertex1=v3, vertex2=v1, multicolor=multicolor)
        edge4 = BGEdge(vertex1=v3, vertex2=v3, multicolor=multicolor)
        self.assertFalse(edge1.is_irregular_edge)
        self.assertTrue(edge2.is_irregular_edge)
        self.assertTrue(edge3.is_irregular_edge)
        self.assertTrue(edge4.is_irregular_edge)

    def test_iter_over_colors_json_ids(self):
        # when multiedge is serialized into json a list of colors in it referenced by their ids
        # the multiplicity of colors has to be preserved
        v1 = BlockVertex("v1")
        v2 = BlockVertex("v2")
        genomes = [self.genome1, self.genome1, self.genome2, self.genome3, self.genome2]
        multicolor = Multicolor(*genomes)
        bgedge = BGEdge(vertex1=v1, vertex2=v2, multicolor=multicolor)
        json_ids = bgedge.colors_json_ids
        self.assertTrue(isinstance(json_ids, list))
        json_ids_list = json_ids
        self.assertEqual(len(json_ids_list), 5)
        ref_json_ids = Counter(genome.json_id for genome in genomes)
        res_json_ids = Counter(json_ids_list)
        self.assertDictEqual(ref_json_ids, res_json_ids)
        # case when color objects are not a BGGenome, but some other hashable object without json_id attribute
        v1 = BlockVertex("v1")
        v2 = BlockVertex("v2")
        colors = ["red", "red", "green", "black", "yellow", "green"]
        multicolor = Multicolor(*colors)
        bgedge = BGEdge(vertex1=v1, vertex2=v2, multicolor=multicolor)
        json_ids = bgedge.colors_json_ids
        self.assertTrue(isinstance(json_ids, list))
        json_ids_list = json_ids
        self.assertEqual(len(json_ids_list), 6)
        ref_json_ids = Counter(hash(genome) for genome in colors)
        res_json_ids = Counter(json_ids_list)
        self.assertDictEqual(ref_json_ids, res_json_ids)
        # case when color objects are mixed objects: BGGenome objects, just hashable, have json_id but not BGGenome
        v1 = BlockVertex("v1")
        v2 = BlockVertex("v2")
        mock1, mock2 = Mock(), Mock()
        mock1.json_id = 5
        mock2.json_id = 6
        colors = [self.genome1, mock1, self.genome2, "black", mock2, self.genome2]
        multicolor = Multicolor(*colors)
        bgedge = BGEdge(vertex1=v1, vertex2=v2, multicolor=multicolor)
        json_ids = bgedge.colors_json_ids
        self.assertTrue(isinstance(json_ids, list))
        json_ids_list = list(json_ids)
        self.assertEqual(len(json_ids_list), 6)
        ref_json_ids = Counter(genome.json_id if hasattr(genome, "json_id") else hash(genome) for genome in colors)
        res_json_ids = Counter(json_ids_list)
        self.assertDictEqual(ref_json_ids, res_json_ids)

    def test_json_serialization(self):
        # simple case of serialization, single color, no multiplicity
        v1, v2 = BlockVertex("v1"), BlockVertex("v2")
        color1 = BGGenome("genome1")
        multicolor = Multicolor(color1)
        edge = BGEdge(vertex1=v1, vertex2=v2, multicolor=multicolor)
        ref_result = {
            "vertex1_id": v1.json_id,
            "vertex2_id": v2.json_id,
            "multicolor": [color1.json_id]
        }
        self.assertDictEqual(edge.to_json(schema_info=False), ref_result)
        # case where multiple colors are present, multiplicity is 1 for every of them
        color2 = BGGenome("genome2")
        multicolor = Multicolor(color1, color2)
        edge = BGEdge(vertex1=v1, vertex2=v2, multicolor=multicolor)
        result = edge.to_json()
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(result["vertex1_id"], v1.json_id)
        self.assertEqual(result["vertex2_id"], v2.json_id)
        self.assertSetEqual(set(result["multicolor"]), {color1.json_id, color2.json_id})
        self.assertEqual(result[BGEdge_JSON_SCHEMA_JSON_KEY], edge.json_schema_name)
        # case where multiple colors are present, multiplicity is both 1 and greater than 1
        color3 = BGGenome("genome3")
        multicolor = Multicolor(color1, color1, color1, color2, color2, color3)
        edge = BGEdge(vertex1=v1, vertex2=v2, multicolor=multicolor)
        result = edge.to_json(schema_info=False)
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(result["vertex1_id"], v1.json_id)
        self.assertEqual(result["vertex2_id"], v2.json_id)
        self.assertSetEqual(set(result["multicolor"]), {color1.json_id, color2.json_id, color3.json_id})
        self.assertDictEqual(Counter(result["multicolor"]), Counter(color.json_id for color in multicolor.multicolors.elements()))
        # weird case when a vertex1/vertex attribute in edge is not an instance of BGVertex
        # and moreover it does not have "json_id" attribute
        edge = BGEdge(vertex1=v1, vertex2=1, multicolor=Multicolor(color1))
        result = edge.to_json()
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(result["vertex1_id"], v1.json_id)
        self.assertEqual(result["vertex2_id"], hash(1))
        self.assertListEqual(result["multicolor"], [color1.json_id])
        self.assertEqual(result[BGEdge_JSON_SCHEMA_JSON_KEY], edge.json_schema_name)

    def test_json_deserialization_default_schema(self):
        # with no scheme is supplied, default scheme for the BGVertex is applied
        # deserialization for vertices and multicolor shall be performed as is, but then it will be resupplied from
        # the overviewing BreakpointGraph
        # correct case no information in json object about schema
        json_object = {
            "vertex1_id": 1,
            "vertex2_id": 2,
            "multicolor": [1, 2, 3, 4]
        }
        result = BGEdge.from_json(data=json_object)
        self.assertTrue(isinstance(result, BGEdge))
        self.assertEqual(result.vertex1, 1)
        self.assertEqual(result.vertex2, 2)
        self.assertListEqual(result.multicolor, [1, 2, 3, 4])
        # correct case with information about json object schema
        # such information about json schema shall be ignored at the BGEdge deserialization level
        json_object = {
            BGEdge_JSON_SCHEMA_JSON_KEY: "dummy_string",
            "vertex1_id": 1,
            "vertex2_id": 2,
            "multicolor": [1, 2, 3, 4]
        }
        result = BGEdge.from_json(data=json_object)
        self.assertTrue(isinstance(result, BGEdge))
        self.assertEqual(result.vertex1, 1)
        self.assertEqual(result.vertex2, 2)
        self.assertListEqual(result.multicolor, [1, 2, 3, 4])
        # incorrect case with at least one vertex id not present in json object
        json_object = {
            "vertex2_id": 2,
            "multicolor": [1, 2, 3, 4]
        }
        with self.assertRaises(ValueError):
            BGEdge.from_json(data=json_object)
        json_object = {
            "vertex1_id": 2,
            "multicolor": [1, 2, 3, 4]
        }
        with self.assertRaises(ValueError):
            BGEdge.from_json(data=json_object)
        # incorrect case with no multicolor present in json object
        json_object = {
            "vertex2_id": 2,
            "vertex1_id": 1,
        }
        with self.assertRaises(ValueError):
            BGEdge.from_json(data=json_object)

    def test_json_deserialization_supplied_schema(self):
        # when a scheme is supplied it shall be used for deserialization
        # correct case no information in json object about schema
        class BGEdgeJSONSchemeDefaultVertex1(BGEdge.BGEdgeJSONSchema):
            def make_object(self, data):
                if "vertex1_json_id" not in data:
                    data["vertex1_json_id"] = 1
                return super().make_object(data)

        json_object = {
            "vertex1_id": 1,
            "vertex2_id": 2,
            "multicolor": [1, 2, 3, 4]
        }
        result = BGEdge.from_json(data=json_object, json_schema_class=BGEdgeJSONSchemeDefaultVertex1)
        self.assertTrue(isinstance(result, BGEdge))
        self.assertEqual(result.vertex1, 1)
        self.assertEqual(result.vertex2, 2)
        self.assertListEqual(result.multicolor, [1, 2, 3, 4])
        # correct case with information about json object schema
        # such information about json schema shall be ignored at the BGEdge deserialization level
        json_object = {
            BGEdge_JSON_SCHEMA_JSON_KEY: "dummy_string",
            "vertex1_id": 1,
            "vertex2_id": 2,
            "multicolor": [1, 2, 3, 4]
        }
        result = BGEdge.from_json(data=json_object, json_schema_class=BGEdgeJSONSchemeDefaultVertex1)
        self.assertTrue(isinstance(result, BGEdge))
        self.assertEqual(result.vertex1, 1)
        self.assertEqual(result.vertex2, 2)
        self.assertListEqual(result.multicolor, [1, 2, 3, 4])
        # incorrect case with at least one vertex id not present in json object
        json_object = {
            "vertex2_id": 2,
            "multicolor": [1, 2, 3, 4]
        }
        result = BGEdge.from_json(data=json_object, json_schema_class=BGEdgeJSONSchemeDefaultVertex1)
        self.assertTrue(isinstance(result, BGEdge))
        self.assertEqual(result.vertex1, 1)
        self.assertEqual(result.vertex2, 2)
        self.assertListEqual(result.multicolor, [1, 2, 3, 4])
        json_object = {
            "vertex1_id": 2,
            "multicolor": [1, 2, 3, 4]
        }
        with self.assertRaises(ValueError):
            BGEdge.from_json(data=json_object)
        # incorrect case with no multicolor present in json object
        json_object = {
            "vertex2_id": 2,
            "vertex1_id": 1,
        }
        with self.assertRaises(ValueError):
            BGEdge.from_json(data=json_object)

if __name__ == '__main__':  # pragma: no cover
    unittest.main()         # pragma: no cover



# -*- coding: utf-8 -*-
from collections import Counter
from copy import deepcopy

__author__ = "Sergey Aganezov"
__email__ = "aganezov(at)gwu.edu"
__status__ = "production"


class Multicolor(object):
    """ Class providing implementation of multi-color notion for edges in :class:`bg.breakpoint_graph.BreakpointGraph`.

    Multi-color is a specific property of edges in Breakpoint Graph combinatorial object which represents similar adjacencies between genomic material in multiple genomes.

    This class supports the following attributes, that carry information colors and their multiplicity of edges in :class:`bg.breakpoint_graph.BreakpointGraph`.

    *    :attr:`Multicolor.multicolors`: a python Counter object which contains information about colors and their multiplicity for a given :class:`Multicolor` instance
    *    :attr:`Multicolor.colors`: a property attribute providing a set of colors in :attr:`Multicolor.multicolors` attribute, hiding information about colors multiplicity

    Main operations:

    *   ``+``, ``-``, ``+=``, ``-=``, ``==``, ``>``, ``>=``, ``<``, ``<=``
    *    :meth:`Multicolor.update`: updates information in :attr:`Multicolor.multicolors` attribute of respective instance
    *    :meth:`Multicolor.merge`: creates a new :class:`Multicolor` object out of a list of provided :class:`Multicolor` objects, gathering respective information about colors and their multiplicity
    *    :meth:`Multicolor.left_merge`: updates respective :class:`Multicolor` instance with information from supplied :class:`Multicolor` object
    *    :meth:`Multicolor.delete`: reduces information in respective instance :attr:`Multicolor.multicolors` attribute by iterating over supplied data
    *    :meth:`Multicolor.similarity_score` computes how similar two supplied :class:`Multicolor` object are
    *    :meth:`Multicolor.split_colors` produces several new instances of :class:`Multicolor` object by splitting information about colors by using provided guidance iterable set-like object
    """

    def __init__(self, *args):
        """ Initialization of :class:`Multicolor` object.

        Initialization is performed by supplied variable number of colors, that respective :class:`Multicolor` object must contain information about Multiplicity of each color is determined by the number of times it occurs as argument in initialization process.

        :param args: variable number of colors to contain information about
        :type args: any hashable python object
        :return: a new instance of :class:`Multicolor`
        :rtype: :class:`Multicolor`
        """
        self.multicolors = Counter(arg for arg in args)

    def update(self, *args):
        """ Updates information about colors and their multiplicity in respective :class:`Multicolor` instance.

        By iterating over supplied arguments each of which should represent a color object, updates information about colors and their multiplicity in current :class:`Multicolor` instance.

        :param args: variable number of colors to add to currently existing multi colors data
        :type args: any hashable python object
        :return: ``None``, performs inplace changes to :attr:`Multicolor.multicolors` attribute
        """
        self.multicolors = self.multicolors + Counter(arg for arg in args)

    @classmethod
    def left_merge(cls, multicolor1, multicolor2):
        """ Updates first supplied :class:`Multicolor` instance with information from second supplied :class:`Multicolor` instance.

        Works as proxy to respective call to private static method :meth:`Multicolor._Multicolor__left_merge` for purposes of inheritance compatibility.

        Accounts for subclassing.

        :param multicolor1: instance to update information in
        :type multicolor1: :class:`Multicolor`
        :param multicolor2: instance to use information for update from
        :type multicolor2: :class:`Multicolor`
        :return: updated first supplied :class:`Multicolor` instance
        :rtype: :class:`Multicolor`
        """
        return cls.__left_merge(multicolor1, multicolor2)

    @classmethod
    def merge(cls, *multicolors):
        """ Produces a new :class:`Multicolor` object resulting from gathering information from all supplied :class:`Multicolor` instances.

        Works as proxy to respective call to private static method :meth:`Multicolor._Multicolor__merge` for purposes of inheritance compatibility.

        :param multicolors: variable number of :class:`Multicolor` objects
        :type multicolors: :class:`Multicolor`
        :return: object containing gathered information from all supplied arguments
        :rtype: :class:`Multicolor`
        """
        return cls.__merge(*multicolors)

    def delete(self, multicolor):
        """ Reduces information :class:`Multicolor` attribute by iterating over supplied colors data.

        Works as proxy to respective call to private static method :meth:`Multicolor._Multicolor__delete` for purposes of inheritance compatibility.

        :param multicolor: information about colors to be deleted from :class:`Multicolor` object
        :type multicolor: any iterable with colors object as entries or :class:`Multicolor`
        :return: ``None``, performs inplace changes
        """
        self.__delete(multicolor)

    def __delete(self, multicolor):
        """ Reduces information :class:`Multicolor` attribute by iterating over supplied colors data.

        In case supplied argument is a :class:`Multicolor` instance, multi-color specific information to de deleted is set to its :attr:`Multicolor.multicolors`.
        In other cases multi-color specific information to de deleted is obtained from iterating over the argument.

        Colors and their multiplicity is reduces with a help of ``-`` method of python Counter object.

        :param multicolor: information about colors to be deleted from :class:`Multicolor` object
        :type multicolor: any iterable with colors object as entries or :class:`Multicolor`
        :return: ``None``, performs inplace changes
        """
        if isinstance(multicolor, Multicolor):
            to_delete = multicolor.multicolors
        else:
            to_delete = Counter(color for color in multicolor)
        self.multicolors = self.multicolors - to_delete

    @classmethod
    def __merge(cls, *multicolors):
        """ Produces a new :class:`Multicolor` object resulting from gathering information from all supplied :class:`Multicolor` instances.

        New :class:`Multicolor` is created and its :attr:`Multicolor.multicolors` attribute is updated with similar attributes of supplied :class:`Multicolor` objects.

        Accounts for subclassing.

        :param multicolors: variable number of :class:`Multicolor` objects
        :type multicolors: :class:`Multicolor`
        :return: object containing gathered information from all supplied arguments
        :rtype: :class:`Multicolor`
        """
        result = cls()
        for multicolor in multicolors:
            result.multicolors = result.multicolors + multicolor.multicolors
        return result

    @staticmethod
    def __left_merge(multicolor1, multicolor2):
        """ Updates first supplied :class:`Multicolor` instance with information from second supplied :class:`Multicolor` instance.

        First supplied instances attribute :attr:`Multicolor.multicolors` is updated with a help of ``+`` method of python Counter object.

        :param multicolor1: instance to update information in
        :type multicolor1: :class:`Multicolor`
        :param multicolor2: instance to use information for update from
        :type multicolor2: :class:`Multicolor`
        :return: updated first supplied :class:`Multicolor` instance
        :rtype: :class:`Multicolor`
        """
        multicolor1.multicolors = multicolor1.multicolors + multicolor2.multicolors
        return multicolor1

    @staticmethod
    def similarity_score(multicolor1, multicolor2):
        """ Computes how similar two :class:`Multicolor` objects are from perspective of information, that they contain.

        Two multicolors are called to be similar if they contain same colors (at least one). Multiplicity of colors is taken into account as well.

        :param multicolor1: first out of two multi-colors to compute similarity between
        :type multicolor1: :class:`Multicolor`
        :param multicolor2: second out of two multi-colors to compute similarity between
        :type multicolor2: :class:`Multicolor`
        :return: the similarity score between two supplied :class:`Multicolor` object
        :rtype: ``int``
        """
        result = 0
        for key, value in multicolor1.multicolors.items():
            if key in multicolor2.multicolors:
                result += min(value, multicolor2.multicolors[key])
        return result

    @classmethod
    def split_colors(cls, multicolor, guidance=None, sorted_guidance=False,
                     account_for_color_multiplicity_in_guidance=True):
        """ Produces several new instances of :class:`Multicolor` object by splitting information about colors by using provided guidance iterable set-like object.

        Guidance is an iterable type of object where each entry has information about groups of colors that has to be separated for current :attr:`Multicolor.multicolors` chunk.
        If no Guidance is provided, single-color guidance of :attr:`Multicolor.multicolors` is created.
        Guidance object is first reversed sorted to iterate over it from larges color set to the smallest one, as small color sets might be subsets of bigger ones, and shall be utilized only if bigger sets didn't help in separating.

        During the first iteration over the guidance information all subsets of :attr:`Multicolor.multicolors` that equal to entries of guidance are recorded.
        During second iteration over remaining of the guidance information, if colors in :attr:`Multicolor.multicolors` form subsets of guidance entries, such instances are recorded.
        After this two iterations, the rest of :attr:`Multicolor.multicolors` is recorded as non-tackled and is recorded on its own.

        Multiplicity of all separated colors in respective chunks is preserved.

        Accounts for subclassing.

        :param multicolor: an instance information about colors in which is to be split
        :type multicolor: :class:`Multicolor`
        :param guidance: information how colors have to be split in current :class:`Multicolor` object
        :type guidance: iterable where each entry is iterable with colors entries
        :param sorted_guidance: a flag, that indicates is sorting of provided guidance is in order
        :return: a list of new :class:`Multicolor` object colors information in which complies with guidance information
        :rtype: ``list`` of :class:`Multicolor` objects
        """
        if guidance is None:
            ###############################################################################################
            #
            # if guidance is not specified, it will be derived from colors in the targeted multicolor
            # initially the multiplicity of colors remains as is
            #
            ###############################################################################################
            guidance = [Multicolor(*(color for _ in range(multicolor.multicolors[color]))) for color in multicolor.colors]
            ###############################################################################################
            #
            # since at this we have a single-colored (maybe with multiplicity greater than 1)
            # we don't need to sort anything, as there will be no overlapping multicolor in guidance
            #
            ###############################################################################################
            sorted_guidance = True
        ###############################################################################################
        #
        # a reference to the targeted multicolor.
        # such reference is created only for the future requirement to access information about original multicolor
        # Is done for the sake of code clarity and consistency.
        #
        ###############################################################################################
        splitting_multicolor = deepcopy(multicolor)
        if not account_for_color_multiplicity_in_guidance:
            ###############################################################################################
            #
            # we need to create a new guidance (even if original is perfect)
            # a new one shall preserve the order of the original, but all multicolors in it
            #   while keeping information about the actual colors itself, shall have multiplicity equal to 1
            #
            ###############################################################################################
            splitting_multicolor = Multicolor(*multicolor.colors)
            colors_guidance = [Multicolor(*multicolor.colors) for multicolor in guidance]
            ###############################################################################################
            #
            # since there might be different multicolors, with the same colors content
            # and they will be changed to same multicolors object, after colors multiplicity adjustment
            # we need, while preserving the order, leave only unique ones in (the first appearance)
            #
            ###############################################################################################
            unique = set()
            guidance = []
            for c_multicolor in colors_guidance:
                if c_multicolor.hashable_representation not in unique:
                    unique.add(c_multicolor.hashable_representation)
                    guidance.append(c_multicolor)
        if not sorted_guidance:
            ###############################################################################################
            #
            # if arguments in function call do not specify explicitly, that the guidance shall be used "as is"
            # it is sorted to put "bigger" multicolors in front, and smaller at the back
            # as bigger multicolor might contain several smaller multicolors from the guidance, but the correct splitting
            # always assumes that the smaller is the splitted result, the better it is
            # and such minimization can be obtained only if the biggest chunk of targeted multicolor are ripped off of it first
            #
            ###############################################################################################
            guidance = sorted({g_multicolor.hashable_representation for g_multicolor in guidance},
                              key=lambda g_multicolor: len(g_multicolor),
                              reverse=True)
            guidance = [Multicolor(*hashed) for hashed in guidance]
        first_run_result = []
        second_run_result = []
        for g_multicolor in guidance:
            ###############################################################################################
            #
            # first we determine which multicolors in guidance are fully present in the multicolor to split
            #   "<=" operator can be read as "is_multi_subset_of"
            # and retrieve as many copies of it from the multicolor, as we can
            # Example:
            #   multicolor has only one color "blue" with multiplicity "4"
            #   in guidance we have multicolor with color "blue" with multiplicity "2"
            #   we must retrieve it fully twice
            #
            ###############################################################################################
            ###############################################################################################
            #
            # empty guidance multicolors shall be ignored, as they have no impact on the splitting algorithm
            #
            ###############################################################################################
            if len(g_multicolor.colors) == 0:
                continue
            while g_multicolor <= splitting_multicolor:
                first_run_result.append(g_multicolor)
                splitting_multicolor -= g_multicolor
        for g_multicolor in guidance:
            ###############################################################################################
            #
            # secondly we determine which multicolors in guidance are partially present in the multicolor
            # NOTE that this is not possible for the case of tree consistent multicolor
            #   as every partially present
            #
            ###############################################################################################
            while len(g_multicolor.intersect(splitting_multicolor).multicolors) > 0:
                second_run_result.append(g_multicolor.intersect(splitting_multicolor))
                splitting_multicolor -= g_multicolor.intersect(splitting_multicolor)
        appendix = splitting_multicolor
        result = deepcopy(first_run_result) + deepcopy(second_run_result) + deepcopy([appendix] if len(appendix.multicolors) > 0 else [])
        if not account_for_color_multiplicity_in_guidance:
            ###############################################################################################
            #
            # if we didn't care for guidance multicolors colors multiplicity, we we splitting a specially created Multicolor
            # based only on the colors content.
            # After this is done, we need to restore the original multiplicity of each color in result multicolors to the
            # count they had in the targeted for splitting multicolor.
            # This is possible since in the case when we do not account for colors multiplicity in guidance, we have
            # splitting_color variable referencing not the supplied multicolor, and thus internal changes are not made to
            # supplied multicolor.
            #
            ###############################################################################################
            for r_multicolor in result:
                for color in r_multicolor.colors:
                    r_multicolor.multicolors[color] = multicolor.multicolors[color]
        return result

    def __sub__(self, other):
        """ Implementation of ``-`` operation for :class:`Multicolor`

        Creates a new :class:`Multicolor` instance by cloning current :class:`Multicolor` object and updating its :attr:`Multicolor.multicolors` attribute information by deleting multi-colors in supplied :class:`Multicolor` object.

        :param other: object, whose multi-color information to subtract form current one
        :type other: :class:`Multicolor`
        :return: new :class:`Multicolor` object, colors in which and their multiplicity result from subtracting of current :attr:`Multicolor.multicolors` and supplied :class:`Multicolor.multicolors` attributes.
        :rtype: :class:`Multicolor`
        :raises: ``TypeError``, if not :class:`Multicolor` instance is supplied
        """
        if not isinstance(other, Multicolor):
            raise TypeError
        result = Multicolor(*(color for color in self.multicolors.elements()))
        result.__delete(other)
        return result

    def __isub__(self, other):
        """ Implementation of ``-`` operation for :class:`Multicolor`

        Updates current :class:`Multicolor` instance by updating its :attr:`Multicolor.multicolors` attribute information by deleting multi-colors in supplied :attr:`Multicolor.multicolors` attribute.
        Utilizes ``-`` method of python Counter

        :param other: object, whose multi-color information to subtract form current one
        :type other: :class:`Multicolor`
        :return: updated current :class:`Multicolor` object
        :rtype: :class:`Multicolor`
        :raises: ``TypeError``, if not :class:`Multicolor` instance is supplied
        """
        if not isinstance(other, Multicolor):
            raise TypeError
        self.multicolors = self.multicolors - other.multicolors
        return self

    def __add__(self, other):
        """ Implementation of ``+`` operation for :class:`Multicolor`

        Invokes a private :meth:`Multicolor._Multicolor__merge` method to implement addition of two :class:`Multicolor` instances.

        :param other: object, whose multi-color information has to be added to current one
        :type other: :class:`Multicolor`
        :return: new :class:`Multicolor` object, colors in which and their multiplicity result from addition of current :attr:`Multicolor.multicolors` and supplied :attr:`Multicolor.multicolors`
        :rtype: :class:`Multicolor`
        :raises: ``TypeError``, if not :class:`Multicolor` instance is provided
        """
        if not isinstance(other, Multicolor):
            raise TypeError
        return Multicolor.__merge(self, other)

    def __iadd__(self, other):
        """ Implementation of ``+=`` operation for :class:`Multicolor`

        Invokes a private :meth:`Multicolor._Multicolor__merge` method to implement addition of two :class:`Multicolor` instances.

        :param other: object, whose multi-color information has to be added to current one
        :type other: :class:`Multicolor`
        :return: new :class:`Multicolor` object, colors in which and their multiplicity result from addition of current :attr:`Multicolor.multicolors` and supplied :attr:`Multicolor.multicolors`
        :rtype: :class:`Multicolor`
        :raises: ``TypeError``, if not :class:`Multicolor` instance is provided
        """
        if not isinstance(other, Multicolor):
            raise TypeError
        return Multicolor.__left_merge(self, other)

    def __eq__(self, other):
        """ Implementation of ``==`` operation for :class:`Multicolor`

        Two :class:`Multicolor` objects are called to be equal if colors that both of them contain and respective colors multiplicity are equal.
        :class:`Multicolor` instance never equals to non-:class:`Multicolor` object.
        Performs :attr:`Multicolor.multicolors` comparison with a help of ``==`` method of python Counter object.

        :param other: an object to compare to
        :type other: any python object
        :return: a flag of equality between current :class:`Multicolor` instance and supplied object
        :rtype: ``Boolean``
        """
        if not isinstance(other, Multicolor):
            return False
        return self.multicolors == other.multicolors

    def __lt__(self, other):
        """ Implementation of ``<`` operation for :class:`Multicolor`

        One :class:`Multicolor` instance is said to be "less than" the other :class:`Multicolor` instance, if it contains less or equal number of colors colors,
        as the other :class:`Multicolor` object does, and multiplicity of all of them is less or equal than in the other multicolor,
        and at least one color has multiplicity less, than in the other multicolor.
        :class:`Multicolor` instance is never less, than non-:class:`Multicolor` object.

        :param other: an object to compare to
        :type other: any python object
        :return: a flag if current :class:`Multicolor` object is less than supplied object
        :rtype: ``Boolean``
        """
        if not isinstance(other, Multicolor):
            return False
        self_keys = self.colors
        other_keys = other.colors
        return all(self.multicolors[key] <= other.multicolors[key] for key in self_keys) and \
               self_keys <= other_keys and \
               any(self.multicolors[key] < other.multicolors[key] for key in self_keys)

    def __le__(self, other):
        """ Implementation of "<=" operation for :class:`Multicolor`

        One :class:`Multicolor` instance is said to be "less or equal than" the other :class:`Multicolor` instance, if it contains less or equal number colors,
        as the other :class:`Multicolor` object does, and multiplicity of all of them is less or equal than in the other multicolor.
        :class:`Multicolor` instance is never less or equal, than non-:class:`Multicolor` object.

        :param other: an object to compare to
        :type other: any python object
        :return: a flag if current :class:`Multicolor` object is less or equal than supplied object
        :rtype: ``Boolean``
        """
        if not isinstance(other, Multicolor):
            return False
        self_keys = self.colors
        other_keys = other.colors
        return all(self.multicolors[key] <= other.multicolors[key] for key in self_keys) and self_keys <= other_keys

    def __gt__(self, other):
        """ Implementation of ``>`` operation for :class:`Multicolor`

        One :class:`Multicolor` instance is said to be "greater than" the other :class:`Multicolor` instance, if it contains greater os equal number of colors,
        as the other :class:`Multicolor` object does, and multiplicity of all of them is greater or equal than in the other multicolor,
        and at least one color has multiplicity greater, than in the other multicolor.
        :class:`Multicolor` instance is never less, than non-:class:`Multicolor` object.

        :param other: an object to compare to
        :type other: any python object
        :return: a flag if current :class:`Multicolor` object is less than supplied object
        :rtype: ``Boolean``
        """
        if not isinstance(other, Multicolor):
            return False
        self_keys = self.colors
        other_keys = other.colors
        return any(self.multicolors[key] > other.multicolors[key] for key in self_keys) and \
               self_keys >= other_keys and \
               all(self.multicolors[key] >= other.multicolors[key] for key in self_keys)

    def __ge__(self, other):
        """ Implementation of ">=" operation for :class:`Multicolor`

        One :class:`Multicolor` instance is said to be "greater than" the other :class:`Multicolor` instance, if it contains greater or equal number of colors,
        as the other :class:`Multicolor` object does, and multiplicity of all of them is greater or equal than in the other multicolor.
        :class:`Multicolor` instance is never less, than non-:class:`Multicolor` object.

        :param other: an object to compare to
        :type other: any python object
        :return: a flag if current :class:`Multicolor` object is greater or equal than supplied object
        :rtype: ``Boolean``
        """
        if not isinstance(other, Multicolor):
            return False
        self_keys = self.colors
        other_keys = other.colors
        return all(self.multicolors[key] >= other.multicolors[key] for key in self_keys) and self_keys >= other_keys

    @property
    def colors(self):
        """ Implements an "attribute" like object to access information about colors only, hiding information about their multiplicity.

        Creates a fresh set object every time is accessed.

        :return: all colors that current :class:`Multicolor` object contains information about.
        :rtype: ``set``
        """
        return set(self.multicolors.keys())

    @property
    def hashable_representation(self):
        """ For a sake of speed check for multicolor presence, each multicolor has a deterministic hashable representation """
        return tuple(sorted(self.multicolors.elements()))

    def __mul__(self, other):
        """ Multicolor can be multiplied by a number and it multiplies multiplicity of each present color respectively

        :param other: an integer multiplier
        :return: a new multicolor object resulted from multiplying each colors multiplicity by the multiplier
        """
        if not isinstance(other, int) or other < 0:
            raise TypeError("Multicolor can be multiplied only by integer values")
        if other == 0:
            return Multicolor()
        result = deepcopy(self)
        for value in result.multicolors:
            result.multicolors[value] *= other
        return result

    def intersect(self, other):
        """ Computes the multiset intersection, between the current Multicolor and the supplied Multicolor

        :param other: another Multicolor object to compute a multiset intersection with
        :return:
        :raise TypeError: an intersection can be computed only between two Multicolor objects
        """
        if not isinstance(other, Multicolor):
            raise TypeError("Multicolor can be intersected only with another Multicolor object")
        intersection_colors_core = self.colors.intersection(other.colors)
        colors_count = {color: min(self.multicolors[color], other.multicolors[color]) for color in intersection_colors_core}
        return Multicolor(*(color for color in colors_count for _ in range(colors_count[color])))


# -*- coding: utf-8 -*-
from unittest.mock import Mock
from bg import Multicolor, BGEdge
from bg.genome import BGGenome
from bg.tree import BGTree, NewickReader, DEFAULT_EDGE_LENGTH

__author__ = "Sergey Aganezov"
__email__ = "aganezov(at)gwu.edu"
__status__ = "production"

import unittest


class BGTreeTestCase(unittest.TestCase):
    def setUp(self):
        # commonly used values during the test cases
        v1, v2, v3, v4, v5 = BGGenome("v1"), BGGenome("v2"), BGGenome("v3"), BGGenome("v4"), BGGenome("v5")
        self.v1 = v1
        self.v2 = v2
        self.v3 = v3
        self.v4 = v4
        self.v5 = v5

    def test_empty_tree_initialization(self):
        # empty tree shall be initialized with no nodes and branches in it
        tree = BGTree()
        self.assertIsNone(tree.root)
        self.assertTrue(tree.is_valid_tree)
        self.assertEqual(len(list(tree.nodes())), 0)
        self.assertEqual(len(list(tree.edges())), 0)
        self.assertListEqual(tree.consistent_multicolors, [Multicolor()])
        self.assertSetEqual(tree.consistent_multicolors_set, {Multicolor().hashable_representation})
        self.assertFalse(tree.account_for_wgd)
        self.assertTrue(tree.multicolors_are_up_to_date)

    def test_add_genome(self):
        # tree support additional of genomes on their own, without edges
        # this is the case for a tree with a single node
        tree = BGTree()
        tree.add_node(BGGenome("genome"))
        self.assertTrue(tree.is_valid_tree)
        self.assertEqual(len(list(tree.nodes())), 1)
        self.assertEqual(len(list(tree.edges())), 0)
        self.assertFalse(tree.multicolors_are_up_to_date)

    def test_edge_length(self):
        # every edge in a tree has a length
        # if no specific length, was specified on edge addition, a default value (1) is stored for this edge
        tree = BGTree()
        tree.add_edge(vertex1=self.v1, vertex2=self.v2, edge_length=5)
        self.assertEqual(tree.edge_length(vertex1=self.v1, vertex2=self.v2), 5)
        self.assertEqual(tree.edge_length(vertex1=self.v2, vertex2=self.v1), 5)
        # edge_length lookup is available only for existing edges, thus both vertices have to be present
        # and an edge between them must exist
        with self.assertRaises(ValueError):
            tree.edge_length(vertex1=self.v1, vertex2=self.v3)
        with self.assertRaises(ValueError):
            tree.edge_length(vertex1=self.v3, vertex2=self.v4)
        with self.assertRaises(ValueError):
            tree.edge_length(vertex1=self.v3, vertex2=self.v4)

    def test_edge_wgd_information(self):
        # every edge in a tree has a number of whole genome duplication (wgd) events assigned to it
        # if no specific count of wgd events was assigned to the edge, a default value (0) is stored for this edge
        tree = BGTree()
        tree.add_edge(vertex1=self.v1, vertex2=self.v2, wgd_events=5)
        tree.add_edge(vertex1=self.v1, vertex2=self.v3)
        self.assertEqual(tree.edge_wgd_count(vertex1=self.v1, vertex2=self.v2), 5)
        self.assertEqual(tree.edge_wgd_count(vertex1=self.v2, vertex2=self.v1), 5)
        self.assertEqual(tree.edge_wgd_count(vertex1=self.v1, vertex2=self.v3), 0)
        self.assertEqual(tree.edge_wgd_count(vertex1=self.v3, vertex2=self.v1), 0)
        # edge_wgd_count lookup is available only for existing edges, thus both vertices have to be present
        # and an edge between them must exist
        with self.assertRaises(ValueError):
            tree.edge_wgd_count(vertex1=self.v1, vertex2=self.v4)
        with self.assertRaises(ValueError):
            tree.edge_wgd_count(vertex1=self.v3, vertex2=self.v4)

    def test_add_edge(self):
        # an edge supports an operation to add a new edge (branch) to the tree
        # if vertices of specified edge were not present in the tree, they are added automatically
        tree = BGTree()
        tree.add_edge(vertex1=self.v1, vertex2=self.v2)
        self.assertTrue(tree.is_valid_tree)
        self.assertEqual(len(list(tree.nodes())), 2)
        self.assertEqual(len(list(tree.edges())), 1)
        self.assertEqual(tree.edge_length(self.v1, self.v2), 1)
        self.assertFalse(tree.multicolors_are_up_to_date)

    def test_add_edge_explicit_edge_length(self):
        # when an edge is added, one can explicitly set its length
        tree = BGTree()
        tree.add_edge(vertex1=self.v1, vertex2=self.v2, edge_length=5)
        self.assertTrue(tree.is_valid_tree)
        self.assertEqual(len(list(tree.nodes())), 2)
        self.assertEqual(len(list(tree.edges())), 1)
        self.assertEqual(tree.edge_length(self.v1, self.v2), 5)
        self.assertFalse(tree.multicolors_are_up_to_date)

    def test_add_edge_explicit_wgd(self):
        # when an edge is added, one can explicitly set the number of whole genome duplication events associated with it
        tree = BGTree()
        tree.add_edge(vertex1=self.v1, vertex2=self.v2, wgd_events=5)
        self.assertEqual(tree.edge_wgd_count(vertex1=self.v1, vertex2=self.v2), 5)
        self.assertEqual(tree.edge_wgd_count(vertex1=self.v2, vertex2=self.v1), 5)
        self.assertFalse(tree.multicolors_are_up_to_date)

    def test_has_edge(self):
        # tree has a O(1) method to check, if an edge exists between two given vertices
        # if vertices are not present in the graph, no exception is raised
        tree = BGTree()
        tree.add_edge(vertex1=self.v1, vertex2=self.v2)
        tree.add_node(self.v3)
        self.assertTrue(tree.has_edge(self.v1, self.v2))
        self.assertTrue(tree.has_edge(self.v2, self.v1))
        self.assertFalse(tree.has_edge(self.v1, self.v3))
        self.assertFalse(tree.has_edge(self.v3, self.v1))
        self.assertFalse(tree.has_edge(self.v3, self.v2))
        self.assertFalse(tree.has_edge(self.v2, self.v3))
        self.assertFalse(tree.has_edge(self.v1, self.v4))
        self.assertFalse(tree.has_edge(self.v4, self.v1))

    def test_has_node(self):
        # tree has a O(1) method to check if a node is present in a tree
        tree = BGTree()
        tree.add_edge(vertex1=self.v1, vertex2=self.v2)
        tree.add_node(self.v3)
        self.assertTrue(tree.has_node(self.v1))
        self.assertTrue(tree.has_node(self.v2))
        self.assertTrue(tree.has_node(self.v3))
        self.assertFalse(tree.has_node(self.v4))

    def test_assign_root(self):
        # a tree has a special pointer to a node, that represents a root of the tree
        # a root can be set to None, though, which shall not in a right-full mind be present in a tree
        tree = BGTree()
        self.assertIsNone(tree.root)
        with self.assertRaises(ValueError):
            tree.root = self.v1
        tree.add_node(self.v1)
        tree.root = self.v1
        self.assertEqual(tree.root, self.v1)
        tree.root = None
        self.assertIsNone(tree.root)
        self.assertFalse(tree.multicolors_are_up_to_date)

    def test_append_tree(self):
        # two trees can be combined together
        # a union of vertices and edges is obtained in it
        tree1 = BGTree()
        tree2 = BGTree()
        tree1.add_edge(vertex1=self.v1, vertex2=self.v2)
        tree2.add_edge(vertex1=self.v2, vertex2=self.v3)
        self.assertTrue(tree1.is_valid_tree)
        self.assertTrue(tree2.is_valid_tree)
        tree1.multicolors_are_up_to_date = True
        tree2.multicolors_are_up_to_date = True
        tree1.append(tree=tree2)
        #####
        self.assertFalse(tree1.multicolors_are_up_to_date)
        self.assertTrue(tree1.is_valid_tree)
        self.assertEqual(len(list(tree1.nodes())), 3)
        self.assertEqual(len(list(tree1.edges())), 2)
        self.assertTrue(tree1.has_edge(vertex1=self.v3, vertex2=self.v2))
        self.assertTrue(tree1.has_edge(vertex1=self.v1, vertex2=self.v2))
        #####
        self.assertTrue(tree2.multicolors_are_up_to_date)
        self.assertTrue(tree2.is_valid_tree)
        self.assertEqual(len(list(tree2.nodes())), 2)
        self.assertEqual(len(list(tree2.edges())), 1)

    def test_set_wgd_count_correct(self):
        # a wgd events count can be set for a specific edge explicitly
        tree = BGTree()
        tree.add_edge(vertex1=self.v1, vertex2=self.v2)
        tree.multicolors_are_up_to_date = True
        self.assertEqual(tree.edge_wgd_count(vertex1=self.v1, vertex2=self.v2), 0)
        tree.set_wgd_count(vertex1=self.v1, vertex2=self.v2, wgd_count=2)
        self.assertFalse(tree.multicolors_are_up_to_date)
        tree.multicolors_are_up_to_date = True
        self.assertEqual(tree.edge_wgd_count(vertex1=self.v1, vertex2=self.v2), 2)
        tree.set_wgd_count(vertex1=self.v2, vertex2=self.v1, wgd_count=3)
        self.assertEqual(tree.edge_wgd_count(vertex1=self.v1, vertex2=self.v2), 3)
        self.assertFalse(tree.multicolors_are_up_to_date)

    def test_set_wgd_count_incorrect(self):
        # only for existing edges such setup if possible
        tree = BGTree()
        tree.multicolors_are_up_to_date = True
        with self.assertRaises(ValueError):
            self.assertEqual(tree.set_wgd_count(vertex1=self.v1, vertex2=self.v2, wgd_count=2), 2)
        # only positive integers are allowed as values for whole genome duplication count
        self.assertTrue(tree.multicolors_are_up_to_date)
        tree.add_edge(vertex1=self.v1, vertex2=self.v2)
        tree.multicolors_are_up_to_date = True
        incorrect_counts = [0.5, "a", (1,), [1]]
        for incorrect_count in incorrect_counts:
            with self.assertRaises(ValueError):
                tree.set_wgd_count(vertex1=self.v1, vertex2=self.v2, wgd_count=incorrect_count)
            self.assertTrue(tree.multicolors_are_up_to_date)

    def test_set_edge_length(self):
        # a length of a specific edge can be set explicitly
        tree = BGTree()
        tree.add_edge(vertex1=self.v1, vertex2=self.v2)
        tree.set_edge_length(vertex1=self.v1, vertex2=self.v2, edge_length=3)
        self.assertEqual(tree.edge_length(vertex1=self.v1, vertex2=self.v2), 3)
        tree.set_edge_length(vertex1=self.v1, vertex2=self.v2, edge_length=5)
        self.assertEqual(tree.edge_length(vertex1=self.v2, vertex2=self.v1), 5)

    def test_set_edge_length_incorrect(self):
        # only for existing edges such setup is possible
        tree = BGTree()
        with self.assertRaises(ValueError):
            tree.set_edge_length(vertex1=self.v1, vertex2=self.v2, edge_length=3)

    def test_get_tree_consistent_multicolors_unrooted_no_wgd_empty(self):
        # in unrooted tree multicolors are tree consistent, if they represent a perfect split of a tree
        # resulting from deletion of a single edge
        # in this case whole genome duplication are not taken into account, as wgd is a directed event
        # and without a root no direction can be chosen
        tree = BGTree()
        tree.account_for_wgd = True
        tree_consistent_multicolors = tree.get_tree_consistent_multicolors(rooted=False, account_for_wgd=False)
        self.assertFalse(tree.account_for_wgd)
        self.assertTrue(tree.multicolors_are_up_to_date)
        self.assertIsInstance(tree_consistent_multicolors, list)
        self.assertListEqual(tree_consistent_multicolors, tree.consistent_multicolors)
        self.assertFalse(tree_consistent_multicolors is tree.consistent_multicolors)
        for obtained_mc, stored_mc in zip(tree_consistent_multicolors, tree.consistent_multicolors):
            self.assertFalse(obtained_mc is stored_mc)
        self.assertSetEqual({mc.hashable_representation for mc in tree_consistent_multicolors},
                            tree.consistent_multicolors_set)
        self.assertEqual(len(tree_consistent_multicolors), 1)
        self.assertIn(Multicolor(), tree_consistent_multicolors)

    def test_get_tree_consistent_multicolors_unrooted_no_wgd_correct(self):
        # if `rooted` argument is set to `False`, then regardless of `tree.root` value outcome shall be the same
        tree = NewickReader.from_string(data_string="(((v1, v2), v3),(v4, v5));")
        for root in [self.v1, self.v2, self.v3, self.v4, self.v5, "1", "2", "3", "4", None]:
            tree.root = root
            self.assertFalse(tree.multicolors_are_up_to_date)
            tree_consistent_multicolors = tree.get_tree_consistent_multicolors(rooted=False, account_for_wgd=False)
            self.assertTrue(tree.multicolors_are_up_to_date)
            self.assertIsInstance(tree_consistent_multicolors, list)
            self.assertListEqual(tree_consistent_multicolors, tree.consistent_multicolors)
            self.assertFalse(tree_consistent_multicolors is tree.consistent_multicolors)
            for obtained_mc, stored_mc in zip(tree_consistent_multicolors, tree.consistent_multicolors):
                self.assertFalse(obtained_mc is stored_mc)
            self.assertSetEqual({mc.hashable_representation for mc in tree_consistent_multicolors},
                                tree.consistent_multicolors_set)
            self.assertEqual(len(tree_consistent_multicolors), 16)
            ref_tree_consistent_multicolors = [
                Multicolor(), Multicolor(self.v1, self.v2, self.v3, self.v4, self.v5),
                Multicolor(self.v1), Multicolor(self.v2, self.v3, self.v4, self.v5),
                Multicolor(self.v2), Multicolor(self.v1, self.v3, self.v4, self.v5),
                Multicolor(self.v3), Multicolor(self.v1, self.v2, self.v4, self.v5),
                Multicolor(self.v4), Multicolor(self.v1, self.v2, self.v3, self.v5),
                Multicolor(self.v5), Multicolor(self.v1, self.v2, self.v3, self.v4),
                Multicolor(self.v1, self.v2), Multicolor(self.v3, self.v4, self.v5),
                Multicolor(self.v4, self.v5), Multicolor(self.v1, self.v2, self.v3),
            ]
            for multicolor in ref_tree_consistent_multicolors:
                self.assertIn(multicolor, tree_consistent_multicolors)

    def test_get_tree_consistent_multicolors_rooted_no_wgd_correct(self):
        # with no account for wgd root specification is irrelevant
        tree = NewickReader.from_string(data_string="(((v1, v2), v3),(v4, v5));")
        for root in [self.v1, self.v2, self.v3, self.v4, self.v4, "1", "2", "3"]:
            tree.root = root
            self.assertFalse(tree.multicolors_are_up_to_date)
            tree_consistent_multicolors = tree.get_tree_consistent_multicolors(rooted=True, account_for_wgd=False)
            self.assertTrue(tree.multicolors_are_up_to_date)
            self.assertIsInstance(tree_consistent_multicolors, list)
            self.assertTrue(tree_consistent_multicolors, tree.consistent_multicolors)
            self.assertFalse(tree_consistent_multicolors is tree.consistent_multicolors)
            for obtained_mc, stored_mc in zip(tree_consistent_multicolors, tree.consistent_multicolors):
                self.assertFalse(obtained_mc is stored_mc)
            self.assertSetEqual({mc.hashable_representation for mc in tree_consistent_multicolors},
                                tree.consistent_multicolors_set)
            self.assertEqual(len(tree_consistent_multicolors), 16)
            ref_tree_consistent_multicolors = [
                Multicolor(), Multicolor(self.v1, self.v2, self.v3, self.v4, self.v5),
                Multicolor(self.v1), Multicolor(self.v2, self.v3, self.v4, self.v5),
                Multicolor(self.v2), Multicolor(self.v1, self.v3, self.v4, self.v5),
                Multicolor(self.v3), Multicolor(self.v1, self.v2, self.v4, self.v5),
                Multicolor(self.v4), Multicolor(self.v1, self.v2, self.v3, self.v5),
                Multicolor(self.v5), Multicolor(self.v1, self.v2, self.v3, self.v4),
                Multicolor(self.v1, self.v2), Multicolor(self.v3, self.v4, self.v5),
                Multicolor(self.v4, self.v5), Multicolor(self.v1, self.v2, self.v3),
            ]
            for multicolor in ref_tree_consistent_multicolors:
                self.assertIn(multicolor, tree_consistent_multicolors)

    def test_get_tree_consistent_multicolors_with_wgd_incorrect(self):
        # if `account_for_wgd` option is set to `True`, `rooted` argument must be set to `True` as well
        tree = BGTree()
        tree.multicolors_are_up_to_date = False
        with self.assertRaises(ValueError):
            tree.get_tree_consistent_multicolors(rooted=False, account_for_wgd=True)
        self.assertFalse(tree.multicolors_are_up_to_date)
        # root of the tree must be not None (can happen if tree is built manually)
        tree = BGTree()
        tree.add_edge(vertex1=self.v1, vertex2=self.v2)
        with self.assertRaises(ValueError):
            tree.get_tree_consistent_multicolors(rooted=False, account_for_wgd=True)
        self.assertFalse(tree.multicolors_are_up_to_date)

    def test_get_tree_consistent_multicolors_with_wgd_correct_non_leaf_root(self):
        # if a root is specified, each wgd even shall duplicate the number of respected genome ends
        tree = NewickReader.from_string(data_string="(((v1, v2), v3),(v4, v5));")
        tree.set_wgd_count(vertex1=self.v1, vertex2="3", wgd_count=1)
        tree.set_wgd_count(vertex1="3", vertex2="2", wgd_count=2)
        tree.root = "1"
        self.assertFalse(tree.multicolors_are_up_to_date)
        tree_consistent_multicolors = tree.get_tree_consistent_multicolors(rooted=True, account_for_wgd=True)
        self.assertTrue(tree.multicolors_are_up_to_date)
        self.assertIsInstance(tree_consistent_multicolors, list)
        self.assertListEqual(tree_consistent_multicolors, tree.consistent_multicolors)
        self.assertFalse(tree_consistent_multicolors is tree.consistent_multicolors)
        for obtained_mc, stored_mc in zip(tree_consistent_multicolors, tree.consistent_multicolors):
            self.assertFalse(obtained_mc is stored_mc)
        self.assertSetEqual({mc.hashable_representation for mc in tree_consistent_multicolors},
                            tree.consistent_multicolors_set)
        self.assertEqual(len(tree_consistent_multicolors), 22)
        overall_multicolor = Multicolor(self.v1) * 8 + Multicolor(self.v2) * 4 + Multicolor(self.v3, self.v4, self.v5)
        ref_tree_consistent_multicolor = [
            Multicolor(), overall_multicolor,
            Multicolor(self.v1), overall_multicolor - Multicolor(self.v1),
            Multicolor(self.v2), overall_multicolor - Multicolor(self.v2),
            Multicolor(self.v3), overall_multicolor - Multicolor(self.v3),
            Multicolor(self.v4), overall_multicolor - Multicolor(self.v4),
            Multicolor(self.v5), overall_multicolor - Multicolor(self.v5),
            Multicolor(self.v1) * 2, overall_multicolor - Multicolor(self.v1) * 2,
            Multicolor(self.v1) * 2 + Multicolor(self.v2), overall_multicolor - Multicolor(self.v1) * 2 - Multicolor(self.v2),
            Multicolor(self.v1) * 4 + Multicolor(self.v2) * 2, overall_multicolor - Multicolor(self.v1) * 4 - Multicolor(self.v2) * 2,
            Multicolor(self.v1) * 8 + Multicolor(self.v2) * 4, overall_multicolor - Multicolor(self.v1) * 8 - Multicolor(self.v2) * 4,
            Multicolor(self.v1) * 8 + Multicolor(self.v2) * 4 + Multicolor(self.v3), Multicolor(self.v4, self.v5),
        ]
        for multicolor in tree_consistent_multicolors:
            self.assertIn(multicolor, ref_tree_consistent_multicolor)

    def test_get_tree_consistent_multicolors_no_wgd_correct_specified_by_argument(self):
        # with rooted tree, if account_for_wgd is set to False, result shall be the same, if it was not rooted
        tree = NewickReader.from_string(data_string="(((v1, v2), v3),(v4, v5));")
        tree.set_wgd_count(vertex1=self.v1, vertex2="3", wgd_count=1)
        tree.set_wgd_count(vertex1="3", vertex2="2", wgd_count=2)
        tree.root = "1"
        self.assertFalse(tree.multicolors_are_up_to_date)
        tree_consistent_multicolors = tree.get_tree_consistent_multicolors(rooted=True, account_for_wgd=False)
        self.assertTrue(tree.multicolors_are_up_to_date)
        self.assertIsInstance(tree_consistent_multicolors, list)
        self.assertListEqual(tree_consistent_multicolors, tree.consistent_multicolors)
        self.assertFalse(tree_consistent_multicolors is tree.consistent_multicolors)
        for obtained_mc, stored_mc in zip(tree_consistent_multicolors, tree.consistent_multicolors):
            self.assertFalse(obtained_mc is stored_mc)
        self.assertEqual(len(tree_consistent_multicolors), 16)
        ref_tree_consistent_multicolors = [
                Multicolor(), Multicolor(self.v1, self.v2, self.v3, self.v4, self.v5),
                Multicolor(self.v1), Multicolor(self.v2, self.v3, self.v4, self.v5),
                Multicolor(self.v2), Multicolor(self.v1, self.v3, self.v4, self.v5),
                Multicolor(self.v3), Multicolor(self.v1, self.v2, self.v4, self.v5),
                Multicolor(self.v4), Multicolor(self.v1, self.v2, self.v3, self.v5),
                Multicolor(self.v5), Multicolor(self.v1, self.v2, self.v3, self.v4),
                Multicolor(self.v1, self.v2), Multicolor(self.v3, self.v4, self.v5),
                Multicolor(self.v4, self.v5), Multicolor(self.v1, self.v2, self.v3),
            ]
        for multicolor in tree_consistent_multicolors:
            self.assertIn(multicolor, ref_tree_consistent_multicolors)

    def test_get_tree_consistent_multicolors_with_wgd_correct_leaf_root(self):
        # when a root is set to a leaf color it is a specific case, as the root branch shall not account
        # for tree consistent multicolor on its own, but rather provide an artificial outgoing branch
        # corresponding to the previous-to-the-tree events and overall multicolor
        tree = NewickReader.from_string(data_string="(((v1, v2), v3),(v4, v5));")
        tree.set_wgd_count(vertex1=self.v1, vertex2="3", wgd_count=1)
        tree.set_wgd_count(vertex1="3", vertex2="2", wgd_count=2)
        tree.root = self.v1
        self.assertFalse(tree.multicolors_are_up_to_date)
        tree_consistent_multicolors = tree.get_tree_consistent_multicolors(rooted=True, account_for_wgd=True)
        self.assertTrue(tree.multicolors_are_up_to_date)
        self.assertIsInstance(tree_consistent_multicolors, list)
        self.assertListEqual(tree_consistent_multicolors, tree.consistent_multicolors)
        self.assertFalse(tree_consistent_multicolors is tree.consistent_multicolors)
        for obtained_mc, stored_mc in zip(tree_consistent_multicolors, tree.consistent_multicolors):
            self.assertFalse(obtained_mc is stored_mc)
        self.assertSetEqual({mc.hashable_representation for mc in tree_consistent_multicolors},
                            tree.consistent_multicolors_set)
        self.assertEqual(len(tree_consistent_multicolors), 21)
        overall_multicolor = Multicolor(self.v1) + Multicolor(self.v2) * 2 + Multicolor(self.v4, self.v5, self.v3) * 8
        ref_tree_consistent_multicolor = [
            Multicolor(), overall_multicolor,
            overall_multicolor - Multicolor(self.v1),
            Multicolor(self.v2), overall_multicolor - Multicolor(self.v2),
            Multicolor(self.v3), overall_multicolor - Multicolor(self.v3),
            Multicolor(self.v4), overall_multicolor - Multicolor(self.v4),
            Multicolor(self.v5), overall_multicolor - Multicolor(self.v5),
            Multicolor(self.v4, self.v5), overall_multicolor - Multicolor(self.v4, self.v5),
            Multicolor(self.v4, self.v5, self.v3), overall_multicolor - Multicolor(self.v4, self.v5, self.v3),
            Multicolor(self.v4, self.v5, self.v3) * 2, overall_multicolor - Multicolor(self.v4, self.v5, self.v3) * 2,
            Multicolor(self.v4, self.v5, self.v3) * 4, overall_multicolor - Multicolor(self.v4, self.v5, self.v3) * 4,
            Multicolor(self.v4, self.v5, self.v3) * 4 + Multicolor(self.v2), overall_multicolor - Multicolor(self.v4, self.v5, self.v3) * 4 - Multicolor(self.v2),
        ]
        for multicolor in tree_consistent_multicolors:
            self.assertIn(multicolor, ref_tree_consistent_multicolor)

    def test_consistent_multicolor_transparent_update_with_change_occurring(self):
        tree = BGTree()
        tcm1 = tree.consistent_multicolors
        self.assertListEqual(tcm1, [Multicolor()])
        tree.add_edge(vertex1=self.v1, vertex2=self.v2)
        tcm2 = tree.consistent_multicolors
        self.assertEqual(len(tcm2), 4)
        for mc in tcm2:
            self.assertIn(mc, [Multicolor(), Multicolor(self.v1), Multicolor(self.v2), Multicolor(self.v1, self.v2)])

    def test_consistent_multicolors_set_transparent_update_with_change_occurring(self):
        tree = BGTree()
        tcm1 = tree.consistent_multicolors_set
        self.assertSetEqual(tcm1, {Multicolor().hashable_representation})
        tree.add_edge(vertex1=self.v1, vertex2=self.v2)
        tcm2 = tree.consistent_multicolors_set
        self.assertEqual(len(tcm2), 4)
        ref = {mc.hashable_representation for mc in [Multicolor(), Multicolor(self.v1), Multicolor(self.v2), Multicolor(self.v1, self.v2)]}
        for mc in tcm2:
            self.assertIn(mc, ref)

    def test_is_multicolor_consistent(self):
        # tests if supplied multicolor complies with tree topology
        ##########################################################################################
        #
        # empty multicolor complies with any tree
        #
        ##########################################################################################
        mc = Multicolor()
        self.assertTrue(BGTree().is_multicolor_consistent(mc))
        ##########################################################################################
        #
        # simple cases
        #
        ##########################################################################################
        tree = NewickReader.from_string("(((v1, v2), v3),(v4, v5));")
        self.assertTrue(tree.is_multicolor_consistent(Multicolor(self.v1)))
        ##########################################################################################
        #
        # a small v1, v2 subtree, still consistent
        #
        ##########################################################################################
        self.assertTrue(tree.is_multicolor_consistent(Multicolor(self.v1, self.v2)))
        ##########################################################################################
        #
        # bigger v1, v2, v3 subtree, still consistent
        #
        ##########################################################################################
        self.assertTrue(tree.is_multicolor_consistent(Multicolor(self.v1, self.v2, self.v3)))
        ##########################################################################################
        #
        # v2, v3 is not a valid subtree (its compliment is two subtrees, instead of one)
        #
        ##########################################################################################
        self.assertFalse(tree.is_multicolor_consistent(Multicolor(self.v2, self.v3)))
        ##########################################################################################
        #
        # if some genomes in multicolor are not present in tree, then multicolor will not be consistent with the tree
        #
        ##########################################################################################
        self.assertFalse(tree.is_multicolor_consistent(Multicolor(self.v1, BGGenome("v6"))))
        ##########################################################################################
        #
        # other cases for a non wgd tree
        #
        ##########################################################################################
        self.assertTrue(tree.is_multicolor_consistent(Multicolor(self.v1, self.v2, self.v3, self.v4)))
        self.assertTrue(tree.is_multicolor_consistent(Multicolor(self.v1, self.v2, self.v3, self.v5)))
        self.assertTrue(tree.is_multicolor_consistent(Multicolor(self.v1, self.v2, self.v4, self.v5)))
        self.assertTrue(tree.is_multicolor_consistent(Multicolor(self.v1, self.v3, self.v4, self.v5)))
        self.assertTrue(tree.is_multicolor_consistent(Multicolor(self.v2, self.v3, self.v4, self.v5)))
        self.assertTrue(tree.is_multicolor_consistent(Multicolor(self.v1, self.v2, self.v3, self.v4, self.v5)))
        self.assertTrue(tree.is_multicolor_consistent(Multicolor(self.v5, self.v4)))
        self.assertTrue(tree.is_multicolor_consistent(Multicolor(self.v3, self.v4, self.v5)))
        self.assertFalse(tree.is_multicolor_consistent(Multicolor(self.v3, self.v5)))
        ##########################################################################################
        #
        # if we have account for wgd, v1, v1 san become a valid subtree
        #
        ##########################################################################################
        tree.set_wgd_count(vertex1=self.v1, vertex2="3", wgd_count=1)
        tree.root = "1"
        tree.account_for_wgd = True
        self.assertTrue(tree.is_multicolor_consistent(Multicolor(self.v1)))
        self.assertTrue(tree.is_multicolor_consistent(Multicolor(self.v1, self.v1)))
        self.assertFalse(tree.is_multicolor_consistent(Multicolor(self.v1, self.v2)))
        self.assertTrue(tree.is_multicolor_consistent(Multicolor(self.v1, self.v1, self.v2)))

    def test_is_bgedge_consistent(self):
        # tests if supplied bgedge has a multicolor that is consistent with tree topology
        v1, v2 = "v1", "v2"
        bgedge = BGEdge(vertex1=v1, vertex2=v2, multicolor=Multicolor())
        ##########################################################################################
        #
        # bgedge with an empty multicolor complies with any tree
        #
        ##########################################################################################
        mc = Multicolor()
        bgedge.multicolor = mc
        self.assertTrue(BGTree().is_bgedge_consistent(bgedge))
        ##########################################################################################
        #
        # simple cases
        #
        ##########################################################################################
        tree = NewickReader.from_string("(((v1, v2), v3),(v4, v5));")
        bgedge.multicolor = Multicolor(self.v1)
        self.assertTrue(tree.is_bgedge_consistent(bgedge))
        ##########################################################################################
        #
        # a small v1, v2 subtree, still consistent
        #
        ##########################################################################################
        bgedge.multicolor = Multicolor(self.v1, self.v2)
        self.assertTrue(tree.is_bgedge_consistent(bgedge))
        ##########################################################################################
        #
        # bigger v1, v2, v3 subtree, still consistent
        #
        ##########################################################################################
        bgedge.multicolor = Multicolor(self.v1, self.v2, self.v3)
        self.assertTrue(tree.is_bgedge_consistent(bgedge))
        ##########################################################################################
        #
        # v2, v3 is not a valid subtree (its compliment is two subtrees, instead of one)
        #
        ##########################################################################################
        bgedge.multicolor = Multicolor(self.v2, self.v3)
        self.assertFalse(tree.is_bgedge_consistent(bgedge))
        ##########################################################################################
        #
        # if some genomes in multicolor are not present in tree, then multicolor will not be consistent with the tree
        #
        ##########################################################################################
        bgedge.multicolor = Multicolor(self.v1, BGGenome("v6"))
        self.assertFalse(tree.is_bgedge_consistent(bgedge))
        ##########################################################################################
        #
        # other cases for a non wgd tree
        #
        ##########################################################################################
        bgedge.multicolor = Multicolor(self.v1, self.v2, self.v3, self.v4)
        self.assertTrue(tree.is_bgedge_consistent(bgedge))
        bgedge.multicolor = Multicolor(self.v1, self.v2, self.v3, self.v5)
        self.assertTrue(tree.is_bgedge_consistent(bgedge))
        bgedge.multicolor = Multicolor(self.v1, self.v2, self.v4, self.v5)
        self.assertTrue(tree.is_bgedge_consistent(bgedge))
        bgedge.multicolor = Multicolor(self.v1, self.v3, self.v4, self.v5)
        self.assertTrue(tree.is_bgedge_consistent(bgedge))
        bgedge.multicolor = Multicolor(self.v2, self.v3, self.v4, self.v5)
        self.assertTrue(tree.is_bgedge_consistent(bgedge))
        bgedge.multicolor = Multicolor(self.v1, self.v2, self.v3, self.v4, self.v5)
        self.assertTrue(tree.is_bgedge_consistent(bgedge))
        bgedge.multicolor = Multicolor(self.v5, self.v4)
        self.assertTrue(tree.is_bgedge_consistent(bgedge))
        bgedge.multicolor = Multicolor(self.v3, self.v4, self.v5)
        self.assertTrue(tree.is_bgedge_consistent(bgedge))
        bgedge.multicolor = Multicolor(self.v3, self.v5)
        self.assertFalse(tree.is_bgedge_consistent(bgedge))
        ##########################################################################################
        #
        # if we have account for wgd, v1, v1 san become a valid subtree
        #
        ##########################################################################################
        tree.set_wgd_count(vertex1=self.v1, vertex2="3", wgd_count=1)
        tree.root = "1"
        tree.account_for_wgd = True
        bgedge.multicolor = Multicolor(self.v1)
        self.assertTrue(tree.is_bgedge_consistent(bgedge))
        bgedge.multicolor = Multicolor(self.v1, self.v1)
        self.assertTrue(tree.is_bgedge_consistent(bgedge))
        bgedge.multicolor = Multicolor(self.v1, self.v2)
        self.assertFalse(tree.is_bgedge_consistent(bgedge))
        bgedge.multicolor = Multicolor(self.v1, self.v1, self.v2)
        self.assertTrue(tree.is_bgedge_consistent(bgedge))


class NewickParserTestCase(unittest.TestCase):
    def test_parse_simple_node_no_edge_length_correct(self):
        # simple node must be a leaf, and all leafs represent genomes
        node_string = "genome"
        node, edge_length = NewickReader.parse_simple_node(node_string)
        self.assertEqual(edge_length, DEFAULT_EDGE_LENGTH)
        self.assertTrue(isinstance(node, BGGenome))
        self.assertEqual(node, BGGenome("genome"))
        node_string = "genome:"
        node, edge_length = NewickReader.parse_simple_node(node_string)
        self.assertEqual(edge_length, DEFAULT_EDGE_LENGTH)
        self.assertTrue(isinstance(node, BGGenome))
        self.assertEqual(node, BGGenome("genome"))

    def test_parse_simple_incorrect_empty_node(self):
        # node name can not be empty
        node_string = ""
        with self.assertRaises(ValueError):
            NewickReader.parse_simple_node(node_string)

    def test_parse_simple_incorrect_multi_semicolon(self):
        node_string = "genome:5:5"
        with self.assertRaises(ValueError):
            NewickReader.parse_simple_node(node_string)

    def test_parse_simple_node_with_edge_length_correct(self):
        # case with correct edge_length `int`
        node_strings = [
            " genome:5",
            "genome :5",
            " genome :5"
        ]
        for node_string in node_strings:
            node, edge_length = NewickReader.parse_simple_node(node_string)
            self.assertEqual(edge_length, 5)
            self.assertTrue(isinstance(node, BGGenome))
            self.assertEqual(node, BGGenome("genome"))
        # case with correct edge_length `double`
        node_strings = [
            "genome:2.1",
            "genome: 2.1",
            "genome:2.1 ",
            "genome: 2.1 "
        ]
        for node_string in node_strings:
            node, edge_length = NewickReader.parse_simple_node(node_string)
            self.assertEqual(edge_length, 2.1)
            self.assertTrue(isinstance(node, BGGenome))
            self.assertEqual(node, BGGenome("genome"))

    def test_parse_simple_node_incorrect_edge_length(self):
        # edge length has to be int/double and nothing else
        incorrectly_formatted_strings = [
            "genome:5.1.1",
            "genome:5a",
            "genome:5/2",
            "genome:test",
            "genome:5.2a"
        ]
        for node_string in incorrectly_formatted_strings:
            with self.assertRaises(ValueError):
                NewickReader.parse_simple_node(node_string)

    def test_separate_into_same_level_nodes_correct(self):
        # empty string shall be parsed into a single entry list with empty string
        data_string = ""
        result_list = NewickReader.separate_into_same_level_nodes(data_string)
        self.assertListEqual(result_list, [""])
        # single node string must be parsed into a single list entry with node info
        data_strings = ["a", "a:5" "a:5.1"]
        for data_string in data_strings:
            self.assertListEqual(NewickReader.separate_into_same_level_nodes(data_string), [data_string])
        # multiple terminal nodes must be parsed into a list of respective information about nodes
        data_string = " a,   b:5, c:2.1,d    "
        ref_list = ["a", "b:5", "c:2.1", "d"]
        result_list = NewickReader.separate_into_same_level_nodes(data_string)
        self.assertListEqual(result_list, ref_list)
        # multiple terminal nodes + non-terminal subtree
        data_string = " a,  b:3.1, (c,(d,e)f)g:1, (h,i)j:2.1   "
        ref_list = ["a", "b:3.1", "(c,(d,e)f)g:1", "(h,i)j:2.1"]
        result_list = NewickReader.separate_into_same_level_nodes(data_string)
        self.assertListEqual(result_list, ref_list)
        # a single subtree as a node
        data_string = "(a,b)"
        ref_list = ["(a,b)"]
        self.assertListEqual(NewickReader.separate_into_same_level_nodes(data_string), ref_list)

    def test_separate_into_same_level_nodes_incorrect(self):
        # some incorrect versions of our newick format, where empty nodes are forbidden
        error_data_string = [
            ",a,b",
            "a,,b",
            "a,b,,",
            "(),,a",
            ",(),a",
            ",(),,"
        ]
        for data_string in error_data_string:
            with self.assertRaises(ValueError):
                NewickReader.separate_into_same_level_nodes(data_string)

    def test_is_non_terminal_subtree(self):
        # newick reader can determine is supplied subtree is actually a terminal node
        data_string = "a"
        self.assertFalse(NewickReader.is_non_terminal_subtree(data_string))
        data_string = "(a,b)"
        self.assertTrue(NewickReader.is_non_terminal_subtree(data_string))
        data_string = "(a,b)c"
        self.assertTrue(NewickReader.is_non_terminal_subtree(data_string))
        data_string = "(a,c):5"
        self.assertTrue(NewickReader.is_non_terminal_subtree(data_string))
        data_string = "(a,c)c:5"
        self.assertTrue(NewickReader.is_non_terminal_subtree(data_string))

    def test_parse_tree_root(self):
        # every root has a tree in a newick format, and it will be assigned to the tree.root pointer
        # this root node name is generated (ass well as other non explicit nodes) on the fly
        data_string = "()a"
        subtree_string, root_string = NewickReader.tree_node_separation(data_string)
        self.assertEqual(subtree_string, "()")
        self.assertEqual(root_string, "a")
        data_string = "()"
        subtree_string, root_string = NewickReader.tree_node_separation(data_string)
        self.assertEqual(subtree_string, "()")
        self.assertEqual(root_string, "")
        data_string = "()a:5"
        subtree_string, root_string = NewickReader.tree_node_separation(data_string)
        self.assertEqual(subtree_string, "()")
        self.assertEqual(root_string, "a:5")
        data_string = "():5"
        subtree_string, root_string = NewickReader.tree_node_separation(data_string)
        self.assertEqual(subtree_string, "()")
        self.assertEqual(root_string, ":5")

    def test_from_string_correct(self):
        # simplest valid test case with a single node
        data_string = "a;"
        tree = NewickReader.from_string(data_string=data_string)
        self.assertEqual(tree.root, BGGenome("a"))
        self.assertEqual(len(list(tree.nodes())), 1)
        self.assertEqual(len(list(tree.edges())), 0)
        # non-terminal nodes, if labeled explicitly are saved as string entries
        data_string = "(a,b)c;"
        tree = NewickReader.from_string(data_string)
        self.assertTrue(isinstance(tree, BGTree))
        self.assertTrue(tree.is_valid_tree)
        self.assertTrue(tree.has_edge(BGGenome("a"), "c"))
        self.assertTrue(tree.has_edge(BGGenome("b"), "c"))
        self.assertEqual(tree.edge_length(BGGenome("a"), "c"), DEFAULT_EDGE_LENGTH)
        self.assertEqual(tree.edge_length(BGGenome("b"), "c"), DEFAULT_EDGE_LENGTH)
        self.assertEqual(tree.edge_wgd_count(BGGenome("a"), "c"), 0)
        self.assertEqual(tree.edge_wgd_count(BGGenome("b"), "c"), 0)
        self.assertEqual(tree.root, "c")
        # non-terminal nodes, if not labeled explicitly are assigned iteratively increased integer values, casted to str
        # since processing is left-to-right, testing for implicitly assigned values can be deterministic
        data_string = "((a,b:5),(c,d):.5);"
        tree = NewickReader.from_string(data_string)
        ga, gb, gc, gd = BGGenome("a"), BGGenome("b"), BGGenome("c"), BGGenome("d")
        self.assertTrue(isinstance(tree, BGTree))
        self.assertTrue(tree.is_valid_tree)
        self.assertTrue(tree.has_edge("2", "1"))
        self.assertTrue(tree.has_edge("3", "1"))
        self.assertTrue(tree.has_edge("2", "1"))
        self.assertTrue(tree.has_edge(ga, "2"))
        self.assertTrue(tree.has_edge(gb, "2"))
        self.assertTrue(tree.has_edge(gc, "3"))
        self.assertTrue(tree.has_edge(gd, "3"))
        self.assertEqual(tree.edge_length(ga, "2"), DEFAULT_EDGE_LENGTH)
        self.assertEqual(tree.edge_length(gb, "2"), 5)
        self.assertEqual(tree.edge_length(gc, "3"), DEFAULT_EDGE_LENGTH)
        self.assertEqual(tree.edge_length(gd, "3"), DEFAULT_EDGE_LENGTH)
        self.assertEqual(tree.edge_length("1", "2"), DEFAULT_EDGE_LENGTH)
        self.assertEqual(tree.edge_length("1", "3"), 0.5)
        # another example
        data_string = "(B:6.0,(A:5.0,C:3.0,E:4.0)Ancestor1:5.0,D:11.0);"
        tree = NewickReader.from_string(data_string)
        ga, gb, gc, gd, ge, anc = BGGenome("A"), BGGenome("B"), BGGenome("C"), BGGenome("D"), BGGenome("E"), "Ancestor1"
        self.assertTrue(isinstance(tree, BGTree))
        self.assertTrue(tree.is_valid_tree)
        self.assertTrue(tree.has_edge(gd, "1"))
        self.assertTrue(tree.has_edge(anc, "1"))
        self.assertTrue(tree.has_edge(gb, "1"))
        self.assertTrue(tree.has_edge(ga, anc))
        self.assertTrue(tree.has_edge(gc, anc))
        self.assertTrue(tree.has_edge(ge, anc))
        self.assertEqual(tree.edge_length(anc, "1"), 5.0)
        self.assertEqual(tree.edge_length(gd, "1"), 11.0)
        self.assertEqual(tree.edge_length(gb, "1"), 6.0)
        self.assertTrue(tree.edge_length(ga, anc), 5.0)
        self.assertTrue(tree.edge_length(gc, anc), 3.0)
        self.assertTrue(tree.edge_length(ge, anc), 4.0)

    def test_from_string_incorrect(self):
        # multiple top level nodes
        data_string = "a,(c,d)b;"
        with self.assertRaises(ValueError):
            NewickReader.from_string(data_string=data_string)
        # empty terminal node
        data_strings = ["(a,);", ";"]
        for data_string in data_strings:
            with self.assertRaises(ValueError):
                NewickReader.from_string(data_string=data_string)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
__author__ = "Sergey Aganezov"
__email__ = "aganezov(at)gwu.edu"
__status__ = "production"

version = "1.4.0"

__all__ = ["bg_io",
           "breakpoint_graph",
           "edge",
           "genome",
           "kbreak",
           "multicolor",
           "tree",
           "vertices"]

from bg.breakpoint_graph import BreakpointGraph
from bg.vertices import BGVertex, BlockVertex, InfinityVertex
from bg.edge import BGEdge
from bg.multicolor import Multicolor
from bg.bg_io import GRIMMReader
from bg.genome import BGGenome
from bg.kbreak import KBreak
from bg.tree import BGTree, NewickReader



# -*- coding: utf-8 -*-
from copy import deepcopy
import itertools
from networkx import Graph
import networkx as nx
from bg import Multicolor
from bg.genome import BGGenome

__author__ = "Sergey Aganezov"
__email__ = "aganezov(at)gwu.edu"
__status__ = "production"

# defines a default edge length in phylogenetic tree
DEFAULT_EDGE_LENGTH = 1
# defines a default number of whole genome duplication events that occur on a branch of a phylogenetic tree
DEFAULT_EDGE_WGD_COUNT = 0


class NewickReader(object):
    """ Class that is designed for reading and parsing a slightly restricted Newick based format for tree representation """

    @classmethod
    def unlabeled_node_names(cls):
        """ A special class based generator for generating names for unlabeled non-terminal nodes """
        for int_value in itertools.count(1):
            yield str(int_value)

    @classmethod
    def parse_node(cls, data_string):
        """ Parses a node in a tree, where such node information is represented in a restricted newick format

        :param data_string: a string with non-terminal node information
        :type data_string: `str`
        :return: name of the node and length of an edge leading to it
        :rtype: `str`, Number
        """
        if ":" in data_string:
            node_name, edge_length_str = data_string.split(":")
            node_name = node_name.strip()
            if len(edge_length_str) == 0:
                edge_length_str = str(DEFAULT_EDGE_LENGTH)
        else:
            node_name = data_string.strip()
            edge_length_str = str(DEFAULT_EDGE_LENGTH)
        edge_length = int(edge_length_str) if "." not in edge_length_str else float(edge_length_str)
        return node_name, edge_length

    @classmethod
    def parse_simple_node(cls, data_string):
        """ Parses terminal node in a tree, where such node information is represented in a restricted newick format

        Checks that node name is not empty, as it is a restriction for Phylogenetic tree,
        where all terminal nodes need to be known

        :param data_string: a string with non-terminal node information
        :type data_string: `str`
        :return: name of the node and length of an edge leading to it
        :rtype: `str`, Number
        :raises: `ValueError` in a case when node name is empty
        """
        node_name, edge_length = cls.parse_node(data_string)
        if len(node_name) == 0:
            raise ValueError("Terminal node can't be empty")
        return BGGenome(node_name), edge_length

    @classmethod
    def separate_into_same_level_nodes(cls, data_string):
        """ Method scans supplied string and determines all subtrees listed in it.

        Performs an additional check to make sure, that there is no empty node listed in supplied subtree string

        :param data_string: a string with subtree information
        :type data_string: `str`
        :return: a list of subtrees (including a terminal subtrees of one node)
        :rtype: `list[str]`
        :raises: ValueError in case when some internal terminal node subtree is empty
        """

        # first cutting point is the beginning of the string
        current_level_separation_comas = [-1]
        overall_result = []
        parenthesis_count = 0
        for cnt, letter in enumerate(data_string):
            ################################################################################################
            #
            # we must be sure, that we count parenthesis, as only top level comas count for separation purposes
            #
            ################################################################################################
            if letter == "(":
                parenthesis_count += 1
            elif letter == ")":
                parenthesis_count -= 1
            elif parenthesis_count == 0 and letter == ",":
                current_level_separation_comas.append(cnt)

        # last cutting point if the end of the string
        current_level_separation_comas.append(len(data_string))
        if len(current_level_separation_comas) == 2 and current_level_separation_comas[1] == 0:
            ################################################################################################
            #
            # if we have exactly two separation points (this is artificial points of start and end of the string)
            # and the second point is present at the second position, we just return an empty string, as nothing is of interest
            #
            ################################################################################################
            return [""]
        for prev_coma_position, current_comma_position in zip(current_level_separation_comas[:-1],
                                                              current_level_separation_comas[1:]):
            ################################################################################################
            #
            # split the string by top level consecutive comas
            #
            #################################################################################################
            overall_result.append(data_string[prev_coma_position + 1: current_comma_position])

        # make sure no flaking white noise is present
        overall_result = [str_data.strip() for str_data in overall_result]
        if any(map(lambda s: len(s) == 0, overall_result)):
            ################################################################################################
            #
            # we restrict out newick tree to contain emtpy non-terminal nodes, as it is
            #   1) useless
            #   2) messes up with tree topology
            #
            ################################################################################################
            raise ValueError("Empty internal node error")
        return overall_result

    @classmethod
    def is_non_terminal_subtree(cls, data_string):
        """ Check is supplied portion of newick formatted string contains a subtree

        :param data_string: a string ro cast a decision about
        :type data_string: `str`
        :return: an answer, is supplied string contains a subtree, or if it is a terminal node
        :rtype: `Boolean`
        """
        return data_string.startswith("(")

    @classmethod
    def tree_node_separation(cls, data_string):
        """ Separates a subtree info from the information, that determines subtree root node, as well as length of a branch to it

        :param data_string: a string with subtree information in it
        :type data_string: `str`
        :return: a pair of strings, separating subtree info with info about subtrees root
        :rtype: `tuple(str, str)`
        """
        last_parenthesis_position = data_string.rfind(")")
        return data_string[:last_parenthesis_position + 1], data_string[last_parenthesis_position + 1:]

    @classmethod
    def from_string(cls, data_string, tree_root=None, name_generator=None):
        """ Main method that reads supplied string (recursive fashion) with newick formatted tree and returns an instance of BGTree

        Performs several top level checks, when called with a full string
        Overall workflow is implemented with a recursion notion, which terminates at the terminal nodes in newick tree
        All errors, that might be encountered during the parsing and processing are bubbling up from the more special methods
            the only exception is a top level check for only a single tree root

        :param data_string: a full newick tree string
        :type data_string: `str`
        :param tree_root: a name of root of the tree, where currently parsed tree shall be attached to (utilized for subsequent calls)
        :type tree_root: `str`
        :param name_generator: a callable that generates names for internal non-terminal nodes, that are left anonymous
        :type name_generator: `callable`
        :return: BGTree that represents the tree encoded in supplied string
        :rtype: `BGTree`
        """
        #################################
        # flanking white noise is always omitted
        data_string = data_string.strip()
        result_tree = BGTree()
        if name_generator is None:
            # NewickReader class has a class based method, that returns a name generator
            # if non specified in the function argument
            name_generator = cls.unlabeled_node_names()
        top_level = ";" in data_string  # semicolon shall be encountered only once during newick tree parsing
        if top_level:
            # if we are on the very top level, we shall just go forward as with a regular subtree and forget about the semicolon
            data_string = data_string[:-1]
        nodes = cls.separate_into_same_level_nodes(data_string=data_string)
        if top_level and len(nodes) > 1:
            # special check, that makes sure that at the very top level tree has only one node, which is a root
            raise ValueError("Top level can not contain more than one root node")
        for current_level_node_str in nodes:
            if cls.is_non_terminal_subtree(current_level_node_str):
                ################################################################################################
                #
                # if we are in a case, when a subtree on current level is not terminal go recursive about it
                #
                ################################################################################################
                subtree_str, node_str = cls.tree_node_separation(data_string=current_level_node_str)
                subtree_str = subtree_str.strip()[1:-1]  # we have parenthesis
                node_name, edge_length = cls.parse_node(node_str)
                if len(node_name) == 0:
                    ################################################################################################
                    #
                    # internal nodes are allowed to be left blank, thus a dummy placer shall be put as their name
                    #
                    ################################################################################################
                    node_name = next(name_generator)
                result_tree.add_node(node_name)
                if top_level:
                    ################################################################################################
                    #
                    # if we are at the very top level, root shall be set to the only node there
                    #
                    ################################################################################################
                    result_tree.root = node_name
                ################################################################################################
                #
                # recursively parse the subtree, specifying that all that subtree shall be appended to subtree root
                #
                ################################################################################################
                subtree = cls.from_string(data_string=subtree_str, tree_root=node_name, name_generator=name_generator)
                ################################################################################################
                #
                # since we've specified a root node name for the subtree parsing, we will get a tree, that has that node in it
                # thus tree appending will be performed against this subtree root node
                #
                ################################################################################################
                result_tree.append(subtree)
                if not top_level:
                    ################################################################################################
                    #
                    # if we are not at the top level, we must attach every current level subtree to the specified root
                    # to get a valid tree at the end
                    #
                    ################################################################################################
                    result_tree.add_edge(vertex1=node_name, vertex2=tree_root, edge_length=edge_length)
            else:
                ################################################################################################
                #
                # we encounter the case, when we hit a terminal level subtree (only a single node there)
                #
                ################################################################################################
                genome, edge_length = cls.parse_simple_node(current_level_node_str)
                if top_level:
                    ################################################################################################
                    #
                    # we are in a very special case situation when a tree is a single node
                    #
                    ################################################################################################
                    result_tree.add_node(genome)
                    result_tree.root = genome
                else:
                    ################################################################################################
                    #
                    # make sure we also attach it to the specified root node
                    #
                    ################################################################################################
                    result_tree.add_edge(vertex1=genome, vertex2=tree_root, edge_length=edge_length)
        return result_tree


class BGTree(object):
    """ Class that is designed to store information about phylogenetic information and relations between multiple genomes

    Class utilizes a networkx Graph object as an internal storage
    This tree can store information about:
        * edge lengths
        * number of whole genome duplication events along certain branches
    """

    # class defined variables that are used as keys when storing edge specific data in the edge attribute dicts
    wgd_events_count_attribute_name = "wgd_events_count"
    edge_length_attribute_name = "edge_length"

    def __init__(self):
        self.__root = None
        self.graph = Graph()
        self.multicolors_are_up_to_date = True
        self.__consistent_multicolors_set = {Multicolor().hashable_representation}
        self.__consistent_multicolors = [Multicolor()]
        self.account_for_wgd = False
        self.__prev_account_for_wgd = False
        self.__prev_rooted = False

    def nodes(self, data=False):
        """ Proxies iteration to the underlying networkx.Graph.nodes iterator

        :param data: a flag to indicate if only nodes are required as iteration result, or associated with them additional data as well
        :type data: `Boolean`
        :return: iterator over all nodes in current tree
        :rtype: iterator
        """
        yield from self.graph.nodes_iter(data=data)

    def edges(self, nbunch=None, data=False):
        """ Proxies iteration to the underlying networkx.Graph.edges iterator

        :param nbunch: an optional parameter to iterate over edges incident to specific set of vertices
        :param data: a flag to indicate if only edges are required as iteration result, or associated with them additional data as well
        :return: iterator over edges in current tree
        :rtype: iterator
        """
        yield from self.graph.edges_iter(nbunch=nbunch, data=data)

    def add_node(self, node):
        """ Adds a new node to current graph
        """
        self.graph.add_node(node)
        self.multicolors_are_up_to_date = False

    def __set_wgd_count(self, vertex1, vertex2, wgd_count):
        """ Explicit method to update a number of wgd events associate with current edge (defined by a pair of vertices)

        :param vertex1: a first vertex in current tree an edge is incident to
        :param vertex2: a second vertex in current tree an edge is incident to
        :param wgd_count: a new value of the number of whole genome duplication events associated with current tree edge
        :type wgd_count: `int`
        :return: nothing, inplace changes
        """
        if not self.has_edge(vertex1=vertex1, vertex2=vertex2):
            raise ValueError("Whole genome duplication count can be assigned only to existing edges")
        if not isinstance(wgd_count, int):
            raise ValueError("Only integer values can be assigned as a whole genome duplication count for tree edges")
        self.graph[vertex1][vertex2][self.wgd_events_count_attribute_name] = wgd_count
        self.multicolors_are_up_to_date = False

    def set_wgd_count(self, vertex1, vertex2, wgd_count):
        """ Subclass safe method for explicit wgd setting, that proxies a call to __set_wgd_count """
        self.__set_wgd_count(vertex1=vertex1, vertex2=vertex2, wgd_count=wgd_count)

    def __set_edge_length(self, vertex1, vertex2, edge_length):
        """ Explicit method to update a length value associated with a specific edge

        :param vertex1: a first vertex in current tree an edge is incident to
        :param vertex2: a second vertex in current tree an edge is incident to
        :param edge_length: a new value of edge length
        :type edge_length: `Number`
        :return: nothing, inplace changes
        :raises: ValueError is the edge to update is not present in current graph
        """
        if not self.has_edge(vertex1=vertex1, vertex2=vertex2):
            raise ValueError("Whole genome duplication count can be assigned only to existing edges")
        self.graph[vertex1][vertex2][self.edge_length_attribute_name] = edge_length

    def set_edge_length(self, vertex1, vertex2, edge_length):
        """ Subclass safe method for explicit edge length setting, that proxies a call to __set_edge_length """
        self.__set_edge_length(vertex1=vertex1, vertex2=vertex2, edge_length=edge_length)

    def add_edge(self, vertex1, vertex2, edge_length=DEFAULT_EDGE_LENGTH, wgd_events=DEFAULT_EDGE_WGD_COUNT):
        """ Adds a enw edge to the current tree with specified characteristics

        Due to networkx.Graph.add_edge method, if vertices, that the edge is specified by are not present in current tree,
        they are added automatically

        Can be used (but not advised against) to update existing edge parameters.

        :param vertex1: a first vertex in current tree an edge is incident to
        :param vertex2: a second vertex in current tree an edge is incident to
        :param edge_length: a length of specified edge
        :param wgd_events: a number of whole genome duplication event associated with current edge
        :return: nothing, inplace changes
        """
        self.graph.add_edge(u=vertex1, v=vertex2)
        self.__set_wgd_count(vertex1=vertex1, vertex2=vertex2, wgd_count=wgd_events)
        self.__set_edge_length(vertex1=vertex1, vertex2=vertex2, edge_length=edge_length)
        self.multicolors_are_up_to_date = False

    @property
    def is_valid_tree(self):
        """ A property like check, to make sure current tree is actually valid

        A tree is called valid if it is a connected graph, in which the number_of_nodes equal to the (number_of_edges + 1)
        """
        nodes_cnt = self.graph.number_of_nodes()
        edges_cnt = self.graph.number_of_edges()
        if nodes_cnt == 0 and edges_cnt == 0:
            ################################################################################################
            #
            # empty tree is special case of a valid tree
            #
            ################################################################################################
            return True
        else:
            return nodes_cnt == edges_cnt + 1 and nx.is_connected(self.graph)

    def __has_edge(self, vertex1, vertex2):
        return self.graph.has_edge(u=vertex1, v=vertex2)

    def has_edge(self, vertex1, vertex2):
        return self.__has_edge(vertex1=vertex1, vertex2=vertex2)

    def has_node(self, vertex):
        return self.graph.has_node(n=vertex)

    @property
    def root(self):
        """ A property based call for the root pointer in current tree """
        return self.__root

    @root.setter
    def root(self, value):
        """ A setter for root, that allows root setting only for existing node in the tree, or None """
        if value is not None and value not in self.graph:
            raise ValueError("Only existing node can be set as root")
        self.__root = value
        self.multicolors_are_up_to_date = False

    def append(self, tree):
        """ Append a specified tree, to a current one in a fashion of taken union of edges and vertices between two trees

        Updated current tree with new information
        Utilizes a networkx.Graph method to transparently add info with data (if any) from supplied tree to surrent one

        :param tree: a tree instance to incorporate information from
        :return: nothing, inplace changes
        """
        self.graph.add_edges_from(tree.graph.edges_iter(data=True))
        self.multicolors_are_up_to_date = False

    def edge_length(self, vertex1, vertex2):
        """ Returns a length of an edge, if exists, from the current tree

        :param vertex1: a first vertex in current tree an edge is incident to
        :param vertex2: a second vertex in current tree an edge is incident to
        :return: a length of specified by a pair of vertices edge
        :rtype: `Number`
        :raises: ValueError, if requested a length of an edge, that is not present in current tree
        """
        if not self.has_edge(vertex1, vertex2):
            raise ValueError("Specified edge is not present in current Tree")
        return self.graph[vertex1][vertex2].get(self.edge_length_attribute_name, DEFAULT_EDGE_LENGTH)

    def edge_wgd_count(self, vertex1, vertex2):
        """ Returns a number of wgd events associated with an edge, if exists, from the current tree

        :param vertex1: a first vertex in current tree an edge is incident to
        :param vertex2: a second vertex in current tree an edge is incident to
        :return: a number of wgd events associated with an edge of specified by a pair of vertices
        :rtype: `int`
        :raises: ValueError, if requested a length of an edge, that is not present in current tree
        """
        if not self.__has_edge(vertex1, vertex2):
            raise ValueError("Specified edge is not present in current Tree")
        return self.graph[vertex1][vertex2].get(self.wgd_events_count_attribute_name, DEFAULT_EDGE_WGD_COUNT)

    def __vertex_is_leaf(self, vertex):
        """ Check is there is only a single edge going from a node, which makes it a tree leaf """
        return vertex in self.graph and len(list(self.graph.edges(nbunch=vertex))) <= 1

    def __get_tree_consistent_vertex_based_hashable_multicolors(self, vertex, parent, account_for_wgd=True):
        """ Traverses a tree in a recursive fashion and obtains all possible subtrees composition

        A color is called tree consistent, if it corresponds to some connected part of a tree
        If one takes whole genome duplication events into account, then colors have to be taken to respective powers of two,
         when connected subtree is observed.

        :param vertex: a root of traversed tree
        :param parent: a parent to current traversed tree root
        :param account_for_wgd: a flag, weather to take whole genome duplication into account, ot not
        :return: a list of tree consistent colors for supplied tree
        """

        #########################################################################################################
        #
        # traverse all nodes achievable for current subtree root, if they are not roots parent
        #
        #########################################################################################################
        descendants = [(v1, v2) for v1, v2 in self.graph.edges(vertex) if v1 != parent and v2 != parent]
        result = []
        #########################################################################################################
        #
        # we in the terminal node, and we shall add a single color to the result
        # the only exception is when we take whole genome duplication events into account and the overall
        #   traversal has started from a terminal node, then it shall not be added into the result multicolors set
        #
        #########################################################################################################
        if self.__vertex_is_leaf(vertex=vertex):
            if account_for_wgd:
                if parent is not None:
                    result.append(Multicolor(vertex))
            else:
                result.append(Multicolor(vertex))
        if len(descendants) > 0:
            #########################################################################################################
            #
            # if we are in the internal node, we initialize current vertex multicolor as empty one
            # otherwise we say that current vertex multicolor is the vertex itself
            #   (possible when we don't take wgd events into consideration and start from a leaf)
            #
            #########################################################################################################
            current_vertex_multicolor = Multicolor() if not self.__vertex_is_leaf(vertex) else Multicolor(vertex)
            for v1, v2 in descendants:
                #########################################################################################################
                #
                # recursively iterate over all descending form current subtree root subtrees
                #
                #########################################################################################################
                child_multicolor = self.__get_tree_consistent_vertex_based_hashable_multicolors(vertex=v2, parent=v1,
                                                                                                account_for_wgd=account_for_wgd)
                #########################################################################################################
                #
                # retrieve a number of whole genome duplication events associated with each descenging subtree
                #
                #########################################################################################################
                edge_wgd_count = self.edge_wgd_count(vertex1=v1, vertex2=v2) if account_for_wgd else 0
                #########################################################################################################
                #
                # append result from recursively traversed subtree to current vertex
                #
                #########################################################################################################
                result.extend(child_multicolor)
                for i in range(1, edge_wgd_count + 1):
                    #########################################################################################################
                    #
                    # if we take whole genome duplications into account, we shall append not just a single case of
                    # descendant subtree root multicolor, but all possible powers of two, for each wgd event on respective edge
                    #
                    #########################################################################################################
                    result.append(child_multicolor[-1] * (2 ** i))
                current_vertex_multicolor += child_multicolor[-1] * (2 ** edge_wgd_count)
            result.append(current_vertex_multicolor)
        return result

    def get_tree_consistent_multicolors(self, rooted=True, account_for_wgd=True):
        """ Checks for basic properties of a request for a consistent multicolors, updates internal values if needed and returns a frech copy of a list of tree consistent multicolors

        :param rooted: a flag to indicate if one want to obtain tree consistent multicolor, assuming tree is rooted somewhere
        :param account_for_wgd: a flag to indicate if whole genome duplication events shall have any impact on tree consistency of multicolors
        :return: a deepcopy of a list of tree consistent multicolors
        :raises: ValueError
        """
        if not self.multicolors_are_up_to_date or rooted != self.__prev_rooted or account_for_wgd != self.__prev_account_for_wgd:
            self.__update_consistet_multicolors(rooted=rooted, account_for_wgd=account_for_wgd)
        self.account_for_wgd = self.__prev_account_for_wgd
        return deepcopy(self.consistent_multicolors)

    def __update_consistet_multicolors(self, rooted=True, account_for_wgd=True):
        """ Updates information about tree consistent multicolors for the tree, that might have been changed

        :param rooted: a flag to indicate if one want to obtain tree consistent multicolor, assuming tree is rooted somewhere
        :param account_for_wgd: a flag to indicate if whole genome duplication events shall have any impact on tree consistency of multicolors
        :return: updates internal values
        :raises: ValueError
        """
        #########################################################################################################
        #
        # whole genome duplication events are directed events, and thus tree traversal direction shall be present
        # such direction is denoted by a root
        # if rooted argument if set to False, but whole genome duplications still shall be taken into account,
        #   or the tree root is set to None, correct tree consistent multicolor can be determined and an error is raised
        #
        #########################################################################################################
        if not rooted and account_for_wgd or rooted and self.root is None:
            raise ValueError("Tree consistent colors, that take whole genome duplication into consideration can not "
                             "be constructed on the unrooted tree")
        #########################################################################################################
        #
        # when we don't account for whole genome duplication events we still need to start from some node
        # but it is irrelevant from which node to start, thus a first possible node is determined as a starting point
        #
        #########################################################################################################
        if not account_for_wgd and self.root is None:
            self.root = next(self.nodes())
        #########################################################################################################
        #
        # get all multicolors starting from current root vertex
        #
        #########################################################################################################
        vertex_based_multicolors = self.__get_tree_consistent_vertex_based_hashable_multicolors(vertex=self.root,
                                                                                                parent=None,
                                                                                                account_for_wgd=account_for_wgd)
        result = []
        #########################################################################################################
        #
        # at the point, where recursive processing of multicolors the overall result will correspond to all possible
        # colors in the provided tree, also known as full Multicolor
        #
        #########################################################################################################
        full_multicolor = vertex_based_multicolors[-1]
        #########################################################################################################
        #
        # for each obtained multicolor we need to make sure we also take its complement into the result
        # as a complement of a tree consistent multicolor is also a tree consistent multicolor
        #
        #########################################################################################################
        for multicolor in vertex_based_multicolors:
            result.append(multicolor)
            supplementary = full_multicolor - multicolor
            #########################################################################################################
            #
            # it is important when we get multicolors taking whole genome duplication events into account and are rooted at a leaf
            # not to take that leaf as a tree consistent multicolor
            #
            #########################################################################################################
            if account_for_wgd and self.__vertex_is_leaf(self.root) and supplementary == Multicolor(self.root):
                continue
            result.append(supplementary)
        #########################################################################################################
        #
        # removing all duplications, as some multicolors we obtained along with their supplementary
        #
        #########################################################################################################
        hashed_vertex_tree_consistent_multicolors = {mc.hashable_representation for mc in result}
        self.consistent_multicolors_set = hashed_vertex_tree_consistent_multicolors
        self.consistent_multicolors = [Multicolor(*hashed_multicolor) for hashed_multicolor in
                                       hashed_vertex_tree_consistent_multicolors]
        self.multicolors_are_up_to_date = True
        self.__prev_account_for_wgd = account_for_wgd
        self.__prev_rooted = rooted

    @property
    def consistent_multicolors(self):
        if not self.multicolors_are_up_to_date or self.account_for_wgd != self.__prev_account_for_wgd or (self.root is not None) != self.__prev_rooted:
            self.__update_consistet_multicolors(rooted=self.root is not None, account_for_wgd=self.account_for_wgd)
        self.account_for_wgd = self.__prev_account_for_wgd
        return self.__consistent_multicolors

    @property
    def consistent_multicolors_set(self):
        if not self.multicolors_are_up_to_date or self.account_for_wgd != self.__prev_account_for_wgd or (self.root is not None) != self.__prev_rooted:
            self.__update_consistet_multicolors(rooted=self.root is not None, account_for_wgd=self.account_for_wgd)
        self.account_for_wgd = self.__prev_account_for_wgd
        return self.__consistent_multicolors_set

    @consistent_multicolors.setter
    def consistent_multicolors(self, value):
        self.__consistent_multicolors = value

    @consistent_multicolors_set.setter
    def consistent_multicolors_set(self, value):
        self.__consistent_multicolors_set = value

    def is_multicolor_consistent(self, multicolor):
        return multicolor.hashable_representation in self.consistent_multicolors_set

    def is_bgedge_consistent(self, bgedge):
        return self.is_multicolor_consistent(bgedge.multicolor)
